# -*- coding:utf-8 -*-
import_format = '''# -*- coding:utf-8 -*-
import os
import sys
import datetime
pathname = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, pathname)
sys.path.insert(0, os.path.abspath(os.path.join(pathname, '..')))

from common{common_path}.functions import *
from airflow.models import Variable
dev=''
retry_delay_minutes = 3
'''

param_format = '''
# get your param from airflow web
{param} = Variable.get('{param}')
'''
# 0 -airflow_break_point-
break_point = '''
#airflow_break_point
def S_break_point():
    return 'success'
'''
# 0 - project_name 1 - file name
execute_python_format = '''
def S_{filename}():
    job = os.path.join(ETL_FILE_PATH, '{project_name}/{filename}.py')
    run_python(job,{param},dev=dev)
'''

execute_python_bash_format = '''
def S_{filename}():
    job = os.path.join(ETL_FILE_PATH ,'{project_name}/{filename}.py')
    return 'python %s %s'%(job,{param})
'''

execute_dataxx_format = '''
def S_{filename}():
    job = os.path.join(ETL_FILE_PATH , '{project_name}/{filename}.json')
    para_dict = dict(zip('{param}'.split(','),[{param}]))
    run_dataxx(job,para_dict,dev=dev)
'''

execute_datax_format = '''
def S_{filename}():
    job = os.path.join(ETL_FILE_PATH , '{project_name}/{filename}.sql')
    para_dict = dict(zip('{param}'.split(','),[{param}]))
    run_datax(job, para_dict, dev=dev)
'''

execute_kafkastarrocks_format = '''
def S_{filename}():
    job = os.path.join(ETL_FILE_PATH , '{project_name}/{filename}.sql')
    para_dict = dict(zip('{param}'.split(','),[{param}]))
    kafkaStarrocks(job, para_dict, dev=dev)
'''


# 0 - project_name, 1 - kettle tranform name, 2 - para string
execute_ktr_format = '''
# execute kettle transform
def S_{filename}():
    job=os.path.join(ETL_FILE_PATH , '{project_name}/{filename}.ktr')
    run_kettle(job,{param_str},dev=dev)
'''

# 0 - project_name, 1 - kettle job name, 2 - para string
execute_kjb_format = '''
# execute kettle job
def S_{filename}():
    job=os.path.join(ETL_FILE_PATH , '{project_name}/{filename}.kjb')
    run_kettle(job,{param_str},dev=dev)
'''

# 0 - project_name, 1 - kettle tranform name, 2 - para string
execute_ktr_r_format = '''
# execute kettle transform
def S_{filename}():
    job=os.path.join(ETL_FILE_PATH_R , '{project_name}/{filename}.ktr')
    run_kettle_remote(job,{param_str},dev=dev)
'''

# 0 - project_name, 1 - kettle job name, 2 - para string
execute_kjb_r_format = '''
# execute kettle job
def S_{filename}():
    job=os.path.join(ETL_FILE_PATH_R , '{project_name}/{filename}.kjb')
    run_kettle_remote(job,{param_str},dev=dev)
'''

execute_hdfs_format = '''
# execute kettle with hdfs job
def S_{filename}():
    job=os.path.join(ETL_FILE_PATH , '{project_name}/{filename}.ktr')
    run_kettle(job,{param_str},dev=dev)
    load_hive(job)
'''

# 0 - project_name, 1 - sql script name,
execute_sql_format = '''
# execute db sql script
def S_{filename}():
    path = os.path.join(ETL_FILE_PATH ,'{project_name}/{filename}.sql')
    para_dict = dict(zip('{param}'.split(','),[{param}]))
    run_sql_file(path,'{db}',para_dict,dev=dev)
'''
# remark = sql string
execute_sqlstr_format = '''
# execute db sql script
def S_{filename}():
    para_dict = dict(zip('{param}'.split(','),[{param}]))
    run_sql_str({remark},'{db}',para_dict,dev=dev)
'''

# 0 - sp name, 1 - para string
execute_sp_format = '''
# execute db procedure
def S_{filename}():
    run_sp('{filename}',[{param}],dev=dev)
'''

execute_sqoop_format = '''
# execute sqoop script
def S_{filename}():
    path = os.path.join(ETL_FILE_PATH , '{project_name}/{filename}.sql')
    para_dict = dict(zip('{param}'.split(','),[{param}]))
    run_sqoop(path, para_dict)
'''

execute_hdfsstarrocks_format = '''
# execute starrocks script
def S_{filename}():
    path = os.path.join(ETL_FILE_PATH , '{project_name}/{filename}.sql')
    para_dict = dict(zip('{param}'.split(','),[{param}]))
    run_starrocks(path, para_dict)
'''

execute_sap_format = '''
# execute sap rfc script
def S_{filename}():
    path = os.path.join(ETL_FILE_PATH , '{project_name}/{filename}.json')
    para_dict = dict(zip('{param}'.split(','),[{param}]))
    run_rfc(path, para_dict)
'''

param_str_format = ''' ' "-param:{item_param}={{}}"'.format({item_param}) '''

dag_import_format = '''# -*- coding:utf-8 -*-
import os
import sys
from airflow.operators.python import PythonOperator, BranchPythonOperator
from airflow.operators.trigger_dagrun import TriggerDagRunOperator
from airflow.operators.bash import BashOperator

from airflow import DAG
import airflow
pathname = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, pathname)
sys.path.insert(0, os.path.abspath(os.path.join(pathname, '..')))

# import your jobs from etl_script
from etl_script.{project_name} import *
'''

dag_config_format = '''
# dag config
args = {{
    'owner': '{owner}',
    'start_date': airflow.utils.dates.days_ago({days_ago}),
    'retries': {retries},
    'retry_delay_minutes': retry_delay_minutes,
    {extra_args}
}}
dag = DAG(
    dag_id='{project_name}',
    tags={tags},
    default_args=args,
    schedule_interval='{schedule_interval}',
    catchup=False,
    description = """{description}"""
)
dag.doc_md = """{dag_remark}"""
'''

dag_jobs_format = '''
{filename} = PythonOperator(
    task_id='{filename}',
    python_callable=S_{filename},
    dag=dag
)
{filename}.ui_color = '{color}'
{filename}.driver = '{driver}'
'''

dag_jobs_format_bash = '''
{filename} = BashOperator(
    task_id='{filename}',
    bash_command=S_{filename}(),
    dag=dag
)
{filename}.ui_color = '{color}'
{filename}.driver = '{driver}'
'''

dag_jobs_format_branch = '''
{filename} = BranchPythonOperator(
    task_id='{filename}',
    python_callable={funname},
    dag=dag
)
{filename}.ui_color = '{color}'
{filename}.driver = '{driver}'
'''

dag_jobs_format_diy = '''
{filename} = PythonOperator(
    task_id='{filename}',
    python_callable={funname},
    dag=dag
)
{filename}.ui_color = '{color}'
{filename}.driver = '{driver}'
'''

dag_jobs_format_trigger = '''
{filename} = TriggerDagRunOperator(
    task_id='{filename}',
    trigger_dag_id='{trigger_dag_id}',
    dag=dag
)
{filename}.ui_color = '{color}'
{filename}.driver = '{driver}'
'''

dag_jobs_format_queue = '''
{filename}.queue = 'bg'
'''

# breaking_point
execute_point_format = '''
# decide Continue or terminate
def S_{dagname}():
    point_test('{dagname}','{sleeptime}','{maxtime}')
'''

# check data
execute_validate_format = '''
# execute validate script
def S_{filename}():
    path = os.path.join(ETL_FILE_PATH , '{project_name}/{filename}.sql')
    para_dict = dict(zip('{param}'.split(','),[{param}]))
    validate(path,'{db}',para_dict,dev=dev)
'''

# check dataset
execute_dataset_format = '''
# execute dataset script
def S_{filename}():
    dataset('{filename}','{param}','{remark}',{tolist})
'''

# check dataset
execute_dash_format = '''
# execute dataset script
def S_{filename}():
    refresh_dash('{filename}','{param}')
'''

refresh_tableau_format = '''
# refresh tableau source script
def S_{filename}():
    refreshTableauSource({sourceID})
'''



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import json
import logging
import os
from typing import Any, Optional

from cryptography.fernet import InvalidToken as InvalidFernetToken
from sqlalchemy import Boolean, Column, Integer, String, Text
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session, reconstructor, synonym

from airflow.configuration import ensure_secrets_loaded
from airflow.models.base import ID_LEN, Base
from airflow.models.crypto import get_fernet
from airflow.utils.log.logging_mixin import LoggingMixin
from airflow.utils.log.secrets_masker import mask_secret
from airflow.utils.session import provide_session

log = logging.getLogger()


class Variable(Base, LoggingMixin):
    """
    Variables are a generic way to store and retrieve arbitrary content or settings
    as a simple key value store within Airflow.
    """

    __tablename__ = "variable"
    __NO_DEFAULT_SENTINEL = object()

    id = Column(Integer, primary_key=True)
    key = Column(String(ID_LEN), unique=True)
    _val = Column('val', Text)
    description = Column(Text)
    is_encrypted = Column(Boolean, unique=False, default=False)

    def __init__(self, key=None, val=None, description=None):
        super().__init__()
        self.key = key
        self.val = val
        self.description = description

    @reconstructor
    def on_db_load(self):
        if self._val:
            mask_secret(self.val, self.key)

    def __repr__(self):
        # Hiding the value
        return f'{self.key} : {self._val}'

    def get_val(self):
        """Get Airflow Variable from Metadata DB and decode it using the Fernet Key"""
        if self._val is not None and self.is_encrypted:
            try:
                fernet = get_fernet()
                return fernet.decrypt(bytes(self._val, 'utf-8')).decode()
            except InvalidFernetToken:
                self.log.error("Can't decrypt _val for key=%s, invalid token or value", self.key)
                return None
            except Exception:
                self.log.error("Can't decrypt _val for key=%s, FERNET_KEY configuration missing", self.key)
                return None
        else:
            return self._val

    def set_val(self, value):
        """Encode the specified value with Fernet Key and store it in Variables Table."""
        if value is not None:
            fernet = get_fernet()
            self._val = fernet.encrypt(bytes(value, 'utf-8')).decode()
            self.is_encrypted = fernet.is_encrypted

    @declared_attr
    def val(cls):
        """Get Airflow Variable from Metadata DB and decode it using the Fernet Key"""
        return synonym('_val', descriptor=property(cls.get_val, cls.set_val))

    @classmethod
    def setdefault(cls, key, default, deserialize_json=False):
        """
        Like a Python builtin dict object, setdefault returns the current value
        for a key, and if it isn't there, stores the default value and returns it.

        :param key: Dict key for this Variable
        :type key: str
        :param default: Default value to set and return if the variable
            isn't already in the DB
        :type default: Mixed
        :param deserialize_json: Store this as a JSON encoded value in the DB
            and un-encode it when retrieving a value
        :return: Mixed
        """
        obj = Variable.get(key, default_var=None, deserialize_json=deserialize_json)
        if obj is None:
            if default is not None:
                Variable.set(key, default, serialize_json=deserialize_json)
                return default
            else:
                raise ValueError('Default Value must be set')
        else:
            return obj

    @classmethod
    def get(
        cls,
        key: str,
        default_var: Any = __NO_DEFAULT_SENTINEL,
        deserialize_json: bool = False,
    ) -> Any:
        """
        Gets a value for an Airflow Variable Key

        :param key: Variable Key
        :param default_var: Default value of the Variable if the Variable doesn't exists
        :param deserialize_json: Deserialize the value to a Python dict
        """
        var_val = Variable.get_variable_from_secrets(key=key)
        if var_val is None:
            if default_var is not cls.__NO_DEFAULT_SENTINEL:
                return default_var
            else:
                raise KeyError(f'Variable {key} does not exist')
        else:
            if deserialize_json:
                obj = json.loads(var_val)
                mask_secret(var_val, key)
                return obj
            else:
                mask_secret(var_val, key)
                return var_val

    @classmethod
    @provide_session
    def set(cls, key: str, value: Any, serialize_json: bool = False, session: Session = None):
        """
        Sets a value for an Airflow Variable with a given Key

        :param key: Variable Key
        :param value: Value to set for the Variable
        :param serialize_json: Serialize the value to a JSON string
        :param session: SQL Alchemy Sessions
        """
        env_var_name = "AIRFLOW_VAR_" + key.upper()
        if env_var_name in os.environ:
            log.warning(
                "You have the environment variable %s defined, which takes precedence over reading "
                "from the database. The value will be saved, but to read it you have to delete "
                "the environment variable.",
                env_var_name,
            )
        if serialize_json:
            stored_value = json.dumps(value, indent=2)
        else:
            stored_value = str(value)

        Variable.delete(key, session=session)
        session.add(Variable(key=key, val=stored_value))
        session.flush()

    @classmethod
    @provide_session
    def delete(cls, key: str, session: Session = None) -> int:
        """
        Delete an Airflow Variable for a given key

        :param key: Variable Key
        :param session: SQL Alchemy Sessions
        """
        return session.query(cls).filter(cls.key == key).delete()

    def rotate_fernet_key(self):
        """Rotate Fernet Key"""
        fernet = get_fernet()
        if self._val and self.is_encrypted:
            self._val = fernet.rotate(self._val.encode('utf-8')).decode()

    @staticmethod
    def get_variable_from_secrets(key: str) -> Optional[str]:
        """
        Get Airflow Variable by iterating over all Secret Backends.

        :param key: Variable Key
        :return: Variable Value
        """
        for secrets_backend in ensure_secrets_loaded():
            var_val = secrets_backend.get_variable(key=key)
            if var_val is not None:
                return var_val
        return None



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
A TaskGroup is a collection of closely related tasks on the same DAG that should be grouped
together when the DAG is displayed graphically.
"""
import functools
from inspect import signature
from typing import Callable, Optional, TypeVar, cast

from airflow.utils.task_group import TaskGroup

T = TypeVar("T", bound=Callable)

task_group_sig = signature(TaskGroup.__init__)


def task_group(python_callable: Optional[Callable] = None, *tg_args, **tg_kwargs) -> Callable[[T], T]:
    """
    Python TaskGroup decorator. Wraps a function into an Airflow TaskGroup.
    Accepts kwargs for operator TaskGroup. Can be used to parametrize TaskGroup.

    :param python_callable: Function to decorate
    :param tg_args: Arguments for TaskGroup object
    :type tg_args: list
    :param tg_kwargs: Kwargs for TaskGroup object.
    :type tg_kwargs: dict
    """

    def wrapper(f: T):
        # Setting group_id as function name if not given in kwarg group_id
        if not tg_args and 'group_id' not in tg_kwargs:
            tg_kwargs['group_id'] = f.__name__
        task_group_bound_args = task_group_sig.bind_partial(*tg_args, **tg_kwargs)

        @functools.wraps(f)
        def factory(*args, **kwargs):
            # Generate signature for decorated function and bind the arguments when called
            # we do this to extract parameters so we can annotate them on the DAG object.
            # In addition, this fails if we are missing any args/kwargs with TypeError as expected.
            # Apply defaults to capture default values if set.

            # Initialize TaskGroup with bound arguments
            with TaskGroup(
                *task_group_bound_args.args, add_suffix_on_collision=True, **task_group_bound_args.kwargs
            ):
                # Invoke function to run Tasks inside the TaskGroup
                return f(*args, **kwargs)

        return cast(T, factory)

    if callable(python_callable):
        return wrapper(python_callable)
    return wrapper



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from flask import current_app
from flask_appbuilder.security.sqla.models import User
from sqlalchemy import func

from airflow.api_connexion import security
from airflow.api_connexion.exceptions import NotFound
from airflow.api_connexion.parameters import apply_sorting, check_limit, format_parameters
from airflow.api_connexion.schemas.user_schema import (
    UserCollection,
    user_collection_item_schema,
    user_collection_schema,
)
from airflow.security import permissions


@security.requires_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_USER)])
def get_user(username):
    """Get a user"""
    ab_security_manager = current_app.appbuilder.sm
    user = ab_security_manager.find_user(username=username)
    if not user:
        raise NotFound(title="User not found", detail=f"The User with username `{username}` was not found")
    return user_collection_item_schema.dump(user)


@security.requires_access([(permissions.ACTION_CAN_READ, permissions.RESOURCE_USER)])
@format_parameters({'limit': check_limit})
def get_users(limit, order_by='id', offset=None):
    """Get users"""
    appbuilder = current_app.appbuilder
    session = appbuilder.get_session
    total_entries = session.query(func.count(User.id)).scalar()
    to_replace = {"user_id": "id"}
    allowed_filter_attrs = [
        "user_id",
        'id',
        "first_name",
        "last_name",
        "user_name",
        "email",
        "is_active",
        "role",
    ]
    query = session.query(User)
    query = apply_sorting(query, order_by, to_replace, allowed_filter_attrs)
    users = query.offset(offset).limit(limit).all()

    return user_collection_schema.dump(UserCollection(users=users, total_entries=total_entries))



from flask import request, jsonify

from airflow.api_connexion import security
from airflow.contrib.dagsetup.gen_airflow_dagfile import gen_airflow_dag, rm_airflow_dag, gen_airflow_config, \
    get_etlfile_txt, save_etlfile_txt, save_dataxfile_txt, save_resfile_txt, rm_dataxfile, rm_resfile
from airflow.exceptions import AirflowException
from airflow.security import permissions


@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_DELETE, permissions.RESOURCE_DAG_RUN),
    ]
)
def post_create_dagfile():
    """Create a dag file."""
    params = request.get_json(force=True)
    print(params)
    try:
        msg = gen_airflow_dag(**params)
    except AirflowException as err:
        response = jsonify(error="{}".format(err))
        response.status_code = err.status_code
        return response
    else:
        return jsonify(msg=msg)


@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_DELETE, permissions.RESOURCE_DAG_RUN),
    ]
)
def get_delete_dagfile():
    """delete the dag."""
    try:
        name = request.args.get('name')
        rm_airflow_dag(name)
    except AirflowException as err:
        response = jsonify(error="{}".format(err))
        response.status_code = err.status_code
        return response
    else:
        return jsonify(msg=f'success delete dag file {name}')


@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_DELETE, permissions.RESOURCE_DAG_RUN),
    ]
)
def get_gen_dagconfig():
    """generate the dag config base on etl script."""
    try:
        name = request.args.get('name')
        dag_str = gen_airflow_config(name)
    except AirflowException as err:
        response = jsonify(error="{}".format(err))
        response.status_code = err.status_code
        return response
    else:
        return jsonify(msg=dag_str)


@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_DELETE, permissions.RESOURCE_DAG_RUN),
    ]
)
def post_create_etlfile():
    """save the etl file context."""
    params = request.get_json(force=True)
    print(params)
    try:
        msg = save_etlfile_txt(**params)
    except AirflowException as err:
        response = jsonify(error="{}".format(err))
        response.status_code = err.status_code
        return response
    else:
        return jsonify(msg=msg)


@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_DELETE, permissions.RESOURCE_DAG_RUN),
    ]
)
def get_gen_etlconfig():
    """get the etl file context."""
    try:
        name = request.args.get('name')
        filename = request.args.get('filename')
        driver = request.args.get('driver')
        etl_dict = get_etlfile_txt(name, filename, driver)
    except AirflowException as err:
        response = jsonify(error="{}".format(err))
        response.status_code = err.status_code
        return response
    else:
        return jsonify(**etl_dict)


@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_DELETE, permissions.RESOURCE_DAG_RUN),
    ]
)
def post_create_dataxfile():
    """save the datax file context."""
    params = request.get_json(force=True)
    print(params)
    try:
        msg = save_dataxfile_txt(**params)
    except AirflowException as err:
        response = jsonify(error="{}".format(err))
        response.status_code = err.status_code
        return response
    else:
        return jsonify(msg=msg)

@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_DELETE, permissions.RESOURCE_DAG_RUN),
    ]
)
def post_create_resfile():
    """save the resource file context."""
    params = request.get_json(force=True)
    print(params)
    try:
        msg = save_resfile_txt(**params)
    except AirflowException as err:
        response = jsonify(error="{}".format(err))
        response.status_code = err.status_code
        return response
    else:
        return jsonify(msg=msg)

@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_DELETE, permissions.RESOURCE_DAG_RUN),
    ]
)
def get_delete_resfile():
    """delete the res."""
    try:
        res_id = request.args.get('res_id')
        rm_resfile(res_id)
    except AirflowException as err:
        response = jsonify(error="{}".format(err))
        response.status_code = err.status_code
        return response
    else:
        return jsonify(msg=f'success delete res {res_id}')


@security.requires_access(
    [
        (permissions.ACTION_CAN_READ, permissions.RESOURCE_DAG),
        (permissions.ACTION_CAN_DELETE, permissions.RESOURCE_DAG_RUN),
    ]
)
def get_delete_dataxfile():
    """delete the datax template"""
    try:
        res_id = request.args.get('res_id')
        tp_id = request.args.get('tp_id')
        rm_dataxfile(res_id, tp_id)
    except AirflowException as err:
        response = jsonify(error="{}".format(err))
        response.status_code = err.status_code
        return response
    else:
        return jsonify(msg=f'success delete res {res_id}, datax tp {tp_id}')



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from functools import wraps
from typing import Callable, Optional, Sequence, Tuple, TypeVar, cast

from flask import current_app, flash, redirect, request, url_for
from flask_login import login_user

from airflow.settings import APPKEY

import time
import base64
import hmac

T = TypeVar("T", bound=Callable)

def generate_token(key, expire=3600):
    r'''
        @Args:
            key: str (用户给定的key，需要用户保存以便之后验证token,每次产生token时的key 都可以是同一个key)
            expire: int(最大有效时间，单位为s)
        @Return:
            state: str
    '''
    ts_str = str(time.time() + expire)
    ts_byte = ts_str.encode("utf-8")
    sha1_tshexstr  = hmac.new(key.encode("utf-8"),ts_byte,'sha1').hexdigest()
    token = ts_str+':'+sha1_tshexstr
    b64_token = base64.urlsafe_b64encode(token.encode("utf-8"))
    return b64_token.decode("utf-8")

def certify_token(key, token):
    r'''
        @Args:
            key: str
            token: str
        @Returns:
            boolean
    '''
    try:
        token_str = base64.urlsafe_b64decode(token).decode('utf-8')
        token_list = token_str.split(':')
        if len(token_list) != 2:
            return False
        ts_str = token_list[0]
        if float(ts_str) < time.time():
            # token expired
            return False
        known_sha1_tsstr = token_list[1]
        sha1 = hmac.new(key.encode("utf-8"),ts_str.encode('utf-8'),'sha1')
        calc_sha1_tsstr = sha1.hexdigest()
        if calc_sha1_tsstr != known_sha1_tsstr:
            # token certification failed
            return False
        # token certification success
        return True
    except Exception as e:
        print(str(e.args))
        return False

def has_access(permissions: Optional[Sequence[Tuple[str, str]]] = None) -> Callable[[T], T]:
    """Factory for decorator that checks current user's permissions against required permissions."""

    def requires_access_decorator(func: T):
        @wraps(func)
        def decorated(*args, **kwargs):
            appbuilder = current_app.appbuilder
            if appbuilder.sm.check_authorization(permissions, request.args.get('dag_id', None)):
                return func(*args, **kwargs)
            else:
                if request.method == 'GET' and request.args.get("user") and request.args.get("EmbedToken"):
                    username = request.args.get("user")
                    token = request.args.get("EmbedToken")
                    if certify_token(APPKEY + username, token):
                        user = appbuilder.sm.auth_user_remote_user(username)
                        if user is None:
                            access_denied = 'auth fail, no user setup'
                        else:
                            login_user(user)
                            return redirect(request.args.get("next") or url_for("Airflow.index"))
                    else:
                        access_denied = 'token auth fail'
                else:
                    access_denied = "Access is Denied"
                flash(access_denied, "danger")

            return redirect(
                url_for(
                    appbuilder.sm.auth_view.__class__.__name__ + ".login",
                    next=request.url,
                )
            )

        return cast(T, decorated)

    return requires_access_decorator




# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Callable, Dict, Iterable, List, Optional, Union

from airflow.decorators.python import python_task
from airflow.decorators.python_virtualenv import _virtualenv_task
from airflow.decorators.task_group import task_group  # noqa
from airflow.models.dag import dag  # noqa


class _TaskDecorator:
    def __call__(
        self, python_callable: Optional[Callable] = None, multiple_outputs: Optional[bool] = None, **kwargs
    ):
        """
        Python operator decorator. Wraps a function into an Airflow operator.
        Accepts kwargs for operator kwarg. This decorator can be reused in a single DAG.

        :param python_callable: Function to decorate
        :type python_callable: Optional[Callable]
        :param multiple_outputs: if set, function return value will be
            unrolled to multiple XCom values. List/Tuples will unroll to xcom values
            with index as key. Dict will unroll to xcom values with keys as XCom keys.
            Defaults to False.
        :type multiple_outputs: bool
        """
        return self.python(python_callable=python_callable, multiple_outputs=multiple_outputs, **kwargs)

    @staticmethod
    def python(python_callable: Optional[Callable] = None, multiple_outputs: Optional[bool] = None, **kwargs):
        """
        Python operator decorator. Wraps a function into an Airflow operator.
        Accepts kwargs for operator kwarg. This decorator can be reused in a single DAG.

        :param python_callable: Function to decorate
        :type python_callable: Optional[Callable]
        :param multiple_outputs: if set, function return value will be
            unrolled to multiple XCom values. List/Tuples will unroll to xcom values
            with index as key. Dict will unroll to xcom values with keys as XCom keys.
            Defaults to False.
        :type multiple_outputs: bool
        """
        return python_task(python_callable=python_callable, multiple_outputs=multiple_outputs, **kwargs)

    @staticmethod
    def virtualenv(
        python_callable: Optional[Callable] = None,
        multiple_outputs: Optional[bool] = None,
        requirements: Optional[Iterable[str]] = None,
        python_version: Optional[Union[str, int, float]] = None,
        use_dill: bool = False,
        system_site_packages: bool = True,
        string_args: Optional[Iterable[str]] = None,
        templates_dict: Optional[Dict] = None,
        templates_exts: Optional[List[str]] = None,
        **kwargs,
    ):
        """
        Allows one to run a function in a virtualenv that is
        created and destroyed automatically (with certain caveats).

        The function must be defined using def, and not be
        part of a class. All imports must happen inside the function
        and no variables outside of the scope may be referenced. A global scope
        variable named virtualenv_string_args will be available (populated by
        string_args). In addition, one can pass stuff through op_args and op_kwargs, and one
        can use a return value.
        Note that if your virtualenv runs in a different Python major version than Airflow,
        you cannot use return values, op_args, op_kwargs, or use any macros that are being provided to
        Airflow through plugins. You can use string_args though.

        .. seealso::
            For more information on how to use this operator, take a look at the guide:
            :ref:`howto/operator:PythonVirtualenvOperator`

        :param python_callable: A python function with no references to outside variables,
            defined with def, which will be run in a virtualenv
        :type python_callable: function
        :param multiple_outputs: if set, function return value will be
            unrolled to multiple XCom values. List/Tuples will unroll to xcom values
            with index as key. Dict will unroll to xcom values with keys as XCom keys.
            Defaults to False.
        :type multiple_outputs: bool
        :param requirements: A list of requirements as specified in a pip install command
        :type requirements: list[str]
        :param python_version: The Python version to run the virtualenv with. Note that
            both 2 and 2.7 are acceptable forms.
        :type python_version: Optional[Union[str, int, float]]
        :param use_dill: Whether to use dill to serialize
            the args and result (pickle is default). This allow more complex types
            but requires you to include dill in your requirements.
        :type use_dill: bool
        :param system_site_packages: Whether to include
            system_site_packages in your virtualenv.
            See virtualenv documentation for more information.
        :type system_site_packages: bool
        :param op_args: A list of positional arguments to pass to python_callable.
        :type op_args: list
        :param op_kwargs: A dict of keyword arguments to pass to python_callable.
        :type op_kwargs: dict
        :param string_args: Strings that are present in the global var virtualenv_string_args,
            available to python_callable at runtime as a list[str]. Note that args are split
            by newline.
        :type string_args: list[str]
        :param templates_dict: a dictionary where the values are templates that
            will get templated by the Airflow engine sometime between
            ``__init__`` and ``execute`` takes place and are made available
            in your callable's context after the template has been applied
        :type templates_dict: dict of str
        :param templates_exts: a list of file extensions to resolve while
            processing templated fields, for examples ``['.sql', '.hql']``
        :type templates_exts: list[str]
        """
        return _virtualenv_task(
            python_callable=python_callable,
            multiple_outputs=multiple_outputs,
            requirements=requirements,
            python_version=python_version,
            use_dill=use_dill,
            system_site_packages=system_site_packages,
            string_args=string_args,
            templates_dict=templates_dict,
            templates_exts=templates_exts,
            **kwargs,
        )


task = _TaskDecorator()



import os #line:2
import requests #line:3
import time #line:4
import json #line:5
import re #line:6
import subprocess #line:7
import logging #line:8
try :#line:9
    from kafka import KafkaConsumer ,TopicPartition #line:10
except :#line:11
    logging .warning ('you need pip install kafka-python')#line:12
os .environ ['NLS_LANG']='SIMPLIFIED CHINESE_CHINA.UTF8'#line:14
requests .packages .urllib3 .disable_warnings ()#line:15
from airflow .settings import ETL_FILE_PATH ,KETTLE_HOME ,HIVE_HOME ,P_URL ,DATASET_TOKEN ,REFRESH_TOKEN #line:17
from airflow .utils .email import fun_email ,list_email #line:18
from airflow .common .datax import datax_cmdStr #line:19
_OO0O00O000O0OO0O0 =f'{P_URL}/echart/dataset_api/?token={DATASET_TOKEN}&visitor=Airflow&type='#line:22
_O0O00O0OO0OO0OOOO =f'{P_URL}/echart/refresh_ds/?token={REFRESH_TOKEN}&type='#line:23
class SmartPipError (Exception ):#line:25
    def __init__ (OO000000OOO0O0OO0 ,err ='SmartPip Error'):#line:26
        Exception .__init__ (OO000000OOO0O0OO0 ,err )#line:27
def smart_upload (OOO000OO0O00O0000 ):#line:30
    OO00O0O0O0O00O000 ,OOO00O0OO0000OO0O =os .path .split (OOO000OO0O00O0000 )#line:31
    OOO00O0OO0000OO0O =OOO00O0OO0000OO0O .split ('.')[0 ]#line:32
    O0O00OO0OO00OOOO0 ={"title":OOO00O0OO0000OO0O ,"token":DATASET_TOKEN ,"visitor":"Airflow"}#line:37
    O0OO00000OO000O00 ={'file':open (OOO000OO0O00O0000 ,'rb')}#line:38
    O0O0000O0OO0O000O =f'''{P_URL}/echart/dataset_api/?type=506&visitor=Airflow&token={DATASET_TOKEN}&param={{"uptime":"{time.time()}","filename":"{OOO00O0OO0000OO0O}"}}'''#line:39
    OOOOOO0OOOOO000OO =60 #line:40
    OO00O0O00O00OO00O =requests .post (f'{P_URL}/etl/api/upload_file_api/',files =O0OO00000OO000O00 ,data =O0O00OO0OO00OOOO0 )#line:42
    print (OO00O0O00O00OO00O .status_code )#line:43
    if OO00O0O00O00OO00O .status_code ==200 :#line:44
        OO00O0O00O00OO00O =OO00O0O00O00OO00O .json ()#line:45
    elif OO00O0O00O00OO00O .status_code ==504 :#line:46
        print ('timeout, try waiting...')#line:47
        OO00O0O00O00OO00O ={"result":"error","data":"time out"}#line:48
        for O0O00OO00000OO000 in range (20 ):#line:49
            O00O0000OO0000O0O =requests .get (O0O0000O0OO0O000O ).json ()#line:50
            print (O00O0000OO0000O0O )#line:51
            O0O00OO0OO00OOOO0 =O00O0000OO0000O0O ['data']#line:52
            if len (O0O00OO0OO00OOOO0 )>1 :#line:53
                OO00O0O00O00OO00O ={"result":"success","data":"uploaded"}#line:54
                break #line:55
            time .sleep (OOOOOO0OOOOO000OO )#line:56
    else :#line:57
        OO00O0O00O00OO00O ={"result":"error","data":"some thing wrong"}#line:58
    print (OO00O0O00O00OO00O )#line:59
    if OO00O0O00O00OO00O ['result']=='error':#line:60
        raise SmartPipError ('Upload Error')#line:61
def get_dataset (OOOOO0OOOO000000O ):#line:62
    ""#line:67
    OOOOOOOOOO00O00O0 =requests .get (_OO0O00O000O0OO0O0 +str (OOOOO0OOOO000000O ),verify =False )#line:68
    OOOOOOOOOO00O00O0 =OOOOOOOOOO00O00O0 .json ()#line:69
    return OOOOOOOOOO00O00O0 #line:70
def dataset (OOO00O0000OOO0000 ,O000OO00000O000OO ,OOO00OOO0OOOO00O0 ,tolist =None ):#line:71
    ""#line:78
    OOO000OOOO00000OO =60 *15 #line:79
    OO0O00O00O000000O =3600 *2 #line:80
    OOO000O0OO0O0O0OO =''#line:81
    try :#line:82
        while True :#line:83
            O0OOOO0OOOOOOOOOO =requests .get (_OO0O00O000O0OO0O0 +O000OO00000O000OO ,verify =False )#line:84
            O0OOOO0OOOOOOOOOO =O0OOOO0OOOOOOOOOO .json ()#line:85
            OO0O00O0000O0OOOO =O0OOOO0OOOOOOOOOO ['result']#line:86
            O0OOOO0OOOOOOOOOO =O0OOOO0OOOOOOOOOO ['data']#line:87
            if OO0O00O0000O0OOOO =='error':#line:88
                raise Exception (f'{O0OOOO0OOOOOOOOOO}')#line:89
            OOO000O0OO0O0O0OO =',\n'.join ([str (O0O0OOO0O0O0OO00O )for O0O0OOO0O0O0OO00O in O0OOOO0OOOOOOOOOO ])#line:90
            print (f'Dataset: {OOO000O0OO0O0O0OO} ')#line:91
            if OOO00OOO0OOOO00O0 =='e3':#line:92
                if len (O0OOOO0OOOOOOOOOO )<2 :#line:93
                    if OO0O00O00O000000O <=0 :#line:94
                        raise Exception ('超时且数据为空')#line:95
                    else :#line:96
                        time .sleep (OOO000OOOO00000OO )#line:97
                        OO0O00O00O000000O =OO0O00O00O000000O -OOO000OOOO00000OO #line:98
                else :#line:99
                    break #line:100
            else :#line:101
                if len (O0OOOO0OOOOOOOOOO )>1 :#line:102
                    if OOO00OOO0OOOO00O0 =='e1':#line:103
                        raise Exception ('有异常数据')#line:104
                    elif OOO00OOO0OOOO00O0 =='e2':#line:105
                        list_email (f'Info_{OOO00O0000OOO0000}',f'{OOO00O0000OOO0000}-Dataset Status',O0OOOO0OOOOOOOOOO ,to_list =tolist )#line:106
                else :#line:107
                    if OOO00OOO0OOOO00O0 not in ['info','e1']:#line:108
                        OOO000O0OO0O0O0OO ='数据为空'#line:109
                        raise Exception (OOO000O0OO0O0O0OO )#line:110
                break #line:111
    except Exception as O00O00OOO000OOO00 :#line:112
        fun_email (f'{OOO00O0000OOO0000}-执行Dataset校验出错',OOO000O0OO0O0O0OO ,to_list =tolist )#line:113
        raise SmartPipError (str (O00O00OOO000OOO00 .args ))#line:114
def refresh_dash (OO0OOO0O00000000O ,O0O0O000OOO0O000O ):#line:115
    ""#line:118
    try :#line:119
        O0OO00OO00OOOOOO0 =requests .get (f'{_O0O00O0OO0OO0OOOO}{O0O0O000OOO0O000O}',verify =False )#line:120
        O0OO00OO00OOOOOO0 =O0OO00OO00OOOOOO0 .json ()#line:121
        print (O0OO00OO00OOOOOO0 )#line:122
        OO00OO0OOO0O0OOO0 =O0OO00OO00OOOOOO0 ['status']#line:123
        if OO00OO0OOO0O0OOO0 !=200 :#line:124
            raise SmartPipError ('refresh_dash')#line:125
    except Exception as O0OO000OOO0O00OO0 :#line:126
        fun_email (f'{OO0OOO0O00000000O}-执行re出错',str (O0OO000OOO0O00OO0 .args ))#line:127
        raise SmartPipError (str (O0OO000OOO0O00OO0 .args ))#line:128
def run_bash (O00OO0O0OOOOO00OO ):#line:131
    O0OO0OO0OO0OO00O0 =''#line:132
    OO0O00OOOO00O0O00 =subprocess .Popen (O00OO0O0OOOOO00OO ,stdout =subprocess .PIPE ,stderr =subprocess .STDOUT ,shell =True ,cwd =ETL_FILE_PATH )#line:133
    print ('PID:',OO0O00OOOO00O0O00 .pid )#line:134
    for OO0O00O00O000OOO0 in iter (OO0O00OOOO00O0O00 .stdout .readline ,b''):#line:135
        if OO0O00OOOO00O0O00 .poll ()and OO0O00O00O000OOO0 ==b'':#line:136
            break #line:137
        OO0O00O00O000OOO0 =OO0O00O00O000OOO0 .decode (encoding ='utf8')#line:138
        print (OO0O00O00O000OOO0 .rstrip ())#line:139
        O0OO0OO0OO0OO00O0 =O0OO0OO0OO0OO00O0 +OO0O00O00O000OOO0 #line:140
    OO0O00OOOO00O0O00 .stdout .close ()#line:141
    O0OOO0O0OO000OOO0 =OO0O00OOOO00O0O00 .wait ()#line:142
    print ('result code: ',O0OOO0O0OO000OOO0 )#line:143
    return O0OO0OO0OO0OO00O0 ,O0OOO0O0OO000OOO0 #line:144
def run_python (O0O0O0OO000O00OOO ,O00OOOO00000OOOO0 ,dev =''):#line:145
    O00O0O0000O00OO00 =O0O0O0OO000O00OOO .split ('/')#line:146
    _O0OOOOOO0OOOO0OOO ,O0OOO0OO00000O000 =run_bash ('python %s %s'%(O0O0O0OO000O00OOO ,O00OOOO00000OOOO0 ))#line:147
    if O0OOO0OO00000O000 !=0 :#line:148
        fun_email (f'{O00O0O0000O00OO00[-2]}/{O00O0O0000O00OO00[-1]}出错','python error')#line:149
        raise Exception ('error')#line:150
def run_dataxx (O000O0OOOO0OO000O ,O0OO0OOO00OOO0OOO ,dev =''):#line:154
    O00O000000OOOO00O =O000O0OOOO0OO000O .split ('/')#line:155
    if O0OO0OOO00OOO0OOO :#line:156
        O0OOO0O000O00OOOO =[f'-D{O0OO0OOOO0O000OOO}:{O0O00OOO0O00O0OO0}'for O0OO0OOOO0O000OOO ,O0O00OOO0O00O0OO0 in O0OO0OOO00OOO0OOO .items ()]#line:157
        OOO000O0O000O00O0 =' '.join (O0OOO0O000O00OOOO )#line:158
        O0000OOO00OOOO0O0 =[f'-p"{OOO000O0O000O00O0}"',O000O0OOOO0OO000O ]#line:159
    else :#line:160
        O0000OOO00OOOO0O0 =[O000O0OOOO0OO000O ]#line:161
    OO00OO0OO0O00OO00 =datax_cmdStr (O0000OOO00OOOO0O0 )#line:162
    _O0000OOOO00O0OO0O ,O000OO00OOO0000OO =run_bash (OO00OO0OO0O00OO00 )#line:163
    if O000OO00OOO0000OO !=0 :#line:164
        fun_email (f'{O00O000000OOOO00O[-2]}/{O00O000000OOOO00O[-1]}出错','datax error')#line:165
        raise Exception ('error')#line:166
def run_datax (O0OO000O00OO00OO0 ,OOOOOOO0O00O0OOO0 ,OOO0O000OOOOO0OO0 ,O000O0OOOO0OO00O0 ,dev =''):#line:167
    with open (O0OO000O00OO00OO0 ,'r',encoding ='utf8')as O00OOO00O0OO00OO0 :#line:168
        O0OOOOOOO00O0O000 =readSqlstr (O00OOO00O0OO00OO0 .read ().strip (),para_dict =O000O0OOOO0OO00O0 )#line:169
    O0OOOOOOO00O0O000 =O0OOOOOOO00O0O000 .split ('##')#line:170
    OO00OO000OO000000 ={}#line:171
    for O0OO00OOOO0OOO0OO in O0OOOOOOO00O0O000 :#line:172
        O0OO000OO0000OO0O =O0OO00OOOO0OOO0OO .find ('=')#line:173
        if O0OO000OO0000OO0O >0 :#line:174
            OO00OO000OO000000 [O0OO00OOOO0OOO0OO [:O0OO000OO0000OO0O ].strip ()]=O0OO00OOOO0OOO0OO [O0OO000OO0000OO0O +1 :].replace ('\n',' ').strip ()#line:175
    O0O0OOOO0OO0OOOOO =OO00OO000OO000000 .keys ()#line:176
    OOOO0OO0O00OO00O0 =OO00OO000OO000000 .pop ('template')if 'template'in O0O0OOOO0OO0OOOOO else 'default'#line:177
    O0O00O0OOO0OOO0O0 =OO00OO000OO000000 .get ('targetColumn')#line:178
    OO00OOOO0O0OOO000 =None #line:179
    if OOOO0OO0O00OO00O0 .endswith ('hdfs'):#line:180
        OO00OOOO0O0OOO000 =OO00OO000OO000000 .pop ('hiveSql')if 'hiveSql'in O0O0OOOO0OO0OOOOO else None #line:182
        if not OO00OOOO0O0OOO000 :#line:183
            OO00OOOO0O0OOO000 =OO00OO000OO000000 .pop ('postSql')if 'postSql'in O0O0OOOO0OO0OOOOO else None #line:184
        if O0O00O0OOO0OOO0O0 :#line:186
            O0O00O0OOO0OOO0O0 =O0O00O0OOO0OOO0O0 .split (',')#line:187
            OOOO00O00OO00O00O =[]#line:188
            for O0OO00OOOO0OOO0OO in O0O00O0OOO0OOO0O0 :#line:189
                if ':'in O0OO00OOOO0OOO0OO :#line:190
                    O0OO00OOOO0OOO0OO =O0OO00OOOO0OOO0OO .split (':')#line:191
                    OOOO00O00OO00O00O .append ({"name":O0OO00OOOO0OOO0OO [0 ].strip (),"type":O0OO00OOOO0OOO0OO [1 ].strip ()})#line:192
                else :#line:193
                    OOOO00O00OO00O00O .append ({"name":O0OO00OOOO0OOO0OO .strip (),"type":"STRING"})#line:194
            OO00OO000OO000000 ['targetColumn']=json .dumps (OOOO00O00OO00O00O )#line:195
    else :#line:196
        if O0O00O0OOO0OOO0O0 :#line:197
            O0O00O0OOO0OOO0O0 =[O0000O0000OOOO00O .strip ()for O0000O0000OOOO00O in O0O00O0OOO0OOO0O0 .split (',')]#line:198
            OO00OO000OO000000 ['targetColumn']=json .dumps (O0O00O0OOO0OOO0O0 )#line:199
        else :#line:200
            OO00OO000OO000000 ['targetColumn']='["*"]'#line:201
        if OOOO0OO0O00OO00O0 .endswith ('starrocks'):#line:203
            if '.'in OO00OO000OO000000 ['targetTable']:#line:204
                OO00OO000OO000000 ['targetDB'],OO00OO000OO000000 ['targetTable']=OO00OO000OO000000 ['targetTable'].split ('.')#line:205
            else :#line:206
                OO00OO000OO000000 ['targetDB']='Test'#line:207
    if 'preSql'in O0O0OOOO0OO0OOOOO :#line:209
        OO00OO000OO000000 ['preSql']=json .dumps (OO00OO000OO000000 ['preSql'].strip ().split (';'))#line:210
    else :#line:211
        OO00OO000OO000000 ['preSql']=''#line:212
    if 'postSql'in O0O0OOOO0OO0OOOOO :#line:213
        OO00OO000OO000000 ['postSql']=json .dumps (OO00OO000OO000000 ['postSql'].strip ().split (';'))#line:214
    else :#line:215
        OO00OO000OO000000 ['postSql']=''#line:216
    O000OO0OO00OO0O0O =O0OO000O00OO00OO0 .split ('/')#line:217
    O00OO0OO00000O00O =O000OO0OO00OO0O0O [-1 ].split ('.')[0 ]#line:218
    with open (os .path .join (OOO0O000OOOOO0OO0 ,'datax','templates',OOOO0OO0O00OO00O0 ),'r')as O00OOO00O0OO00OO0 :#line:219
        O0OO00OOOO000O000 =O00OOO00O0OO00OO0 .read ()#line:220
    O0OO000O00OO00OO0 =os .path .join (OOO0O000OOOOO0OO0 ,'datax',O00OO0OO00000O00O +'.json')#line:221
    with open (O0OO000O00OO00OO0 ,'w',encoding ='utf8')as O00OOO00O0OO00OO0 :#line:222
        O00OOO00O0OO00OO0 .write (readSqlstr (O0OO00OOOO000O000 ,OO00OO000OO000000 ))#line:223
    OOOO0O0O0000O0O0O =datax_cmdStr ([O0OO000O00OO00OO0 ])#line:224
    _O0000000OO00O00OO ,OO00O00O0OO0OO0OO =run_bash (OOOO0O0O0000O0O0O )#line:225
    if OO00O00O0OO0OO0OO !=0 :#line:226
        fun_email (f'{O000OO0OO00OO0O0O[-2]}/{O000OO0OO00OO0O0O[-1]}出错','datax error')#line:227
        raise Exception ('error')#line:228
    if OO00OOOO0O0OOO000 :#line:229
        print (_O0O00000OO000OOO0 (OO00OOOO0O0OOO000 .split (';'),OOOOOOO0O00O0OOO0 ,db_connect ='hive',dev =dev ))#line:230
def readSqlFile (OO0OOOOOOO0O0OO0O ,para_dict =None ):#line:234
    if OO0OOOOOOO0O0OO0O .find ('.sql')<0 :#line:235
        return 'file type error'#line:236
    with open (OO0OOOOOOO0O0OO0O ,'r',encoding ='utf-8')as O00O000OO0000OO0O :#line:237
        OO00O000OO0O000O0 =O00O000OO0000OO0O .read ()#line:238
    O0OOOO000OOOOO0OO =readSqlstr (OO00O000OO0O000O0 ,para_dict )#line:239
    return O0OOOO000OOOOO0OO #line:240
def readSqoopFile (O0OO00O000O00O00O ,para_dict =None ):#line:241
    if not O0OO00O000O00O00O .endswith ('.sql'):#line:242
        return 'file type error'#line:243
    with open (O0OO00O000O00O00O ,'r',encoding ='utf8')as OO0O000OO0000OOOO :#line:244
        OOOOOO00O0OO0O0O0 =OO0O000OO0000OOOO .read ().strip ()#line:245
    OOO00O0O0OOO00000 =re .match (r"/\*(.*?)\*/(.+)",OOOOOO00O0OO0O0O0 ,re .M |re .S )#line:246
    O0O0OOOO0000OOO0O =readSqlstr (OOO00O0O0OOO00000 .group (1 ).strip (),para_dict )#line:247
    O00O00O0OO0OOOOOO =OOO00O0O0OOO00000 .group (2 ).strip ()#line:248
    return O0O0OOOO0000OOO0O ,O00O00O0OO0OOOOOO #line:249
def readSqlstr (O00O0O00O00O00000 ,para_dict =None ):#line:250
    O000OOOO00000OO00 =re .sub (r"(\/\*(.|\n)*?\*\/)|--.*",'',O00O0O00O00O00000 .strip ())#line:251
    if para_dict :#line:252
        for OOO00O0O0O0OOOOOO ,OO0O00OOO0OO00OOO in para_dict .items ():#line:253
            O000OOOO00000OO00 =O000OOOO00000OO00 .replace ('$'+OOO00O0O0O0OOOOOO ,str (OO0O00OOO0OO00OOO ))#line:254
    return O000OOOO00000OO00 #line:255
def run_sql_file (OOO0OO000O00O000O ,OO000O00O00O00000 ,db_connect ='starrocks',para_dict =None ,dev =''):#line:256
    OO0O000O000O00000 =OOO0OO000O00O000O .split ('/')#line:257
    try :#line:258
        OO0O0000O0OOOOO0O =readSqlFile (OOO0OO000O00O000O ,para_dict ).split (';')#line:259
        O000O000O0OO0OOO0 =OO000O00O00O00000 .get (db_connect )#line:260
        if dev :#line:261
            if f'{db_connect}{dev}'in OO000O00O00O00000 .keys ():#line:262
                O000O000O0OO0OOO0 =OO000O00O00O00000 .get (f'{db_connect}{dev}')#line:263
        O000000O0O000O000 =connect_db_execute ().execute_sql_list (OO0O0000O0OOOOO0O ,db_connect ,connect_dict =O000O000O0OO0OOO0 )#line:264
        return O000000O0O000O000 #line:265
    except Exception as O00OOOO000O0OO0O0 :#line:266
        fun_email ('{}/{}执行出错'.format (OO0O000O000O00000 [-2 ],OO0O000O000O00000 [-1 ]),str (O00OOOO000O0OO0O0 .args ))#line:267
        print (O00OOOO000O0OO0O0 .args )#line:268
        raise SmartPipError ('Run SQL Error')#line:269
def _O0O00000OO000OOO0 (OOO0OO000O000O000 ,OOO0O00000OOOOOOO ,db_connect ='starrocks',para_dict =None ,dev =''):#line:270
    try :#line:271
        if isinstance (OOO0OO000O000O000 ,str ):#line:272
            OOO0OO000O000O000 =readSqlstr (OOO0OO000O000O000 ,para_dict ).split (';')#line:273
        OOOOOOOOOO0O00OO0 =OOO0O00000OOOOOOO .get (db_connect )#line:274
        if dev :#line:275
            if f'{db_connect}{dev}'in OOO0O00000OOOOOOO .keys ():#line:276
                OOOOOOOOOO0O00OO0 =OOO0O00000OOOOOOO .get (f'{db_connect}{dev}')#line:277
        OOO00OO0OOOOOOOOO =connect_db_execute ().execute_sql_list (OOO0OO000O000O000 ,db_connect ,connect_dict =OOOOOOOOOO0O00OO0 )#line:278
        return OOO00OO0OOOOOOOOO #line:279
    except Exception as OOOOOO00O0OOOO0O0 :#line:280
        fun_email ('SQL执行出错',f'{OOO0OO000O000O000}{OOOOOO00O0OOOO0O0.args}')#line:281
        print (OOOOOO00O0OOOO0O0 .args )#line:282
        raise SmartPipError ('Run SQL Error')#line:283
def run_kettle (O0O00OOOOO00000O0 ,para_str ='',dev =False ):#line:286
    ""#line:293
    OOO0O00OO0OO0O0OO =O0O00OOOOO00000O0 .split ('/')#line:294
    print ('kettle job start')#line:295
    if '.ktr'in O0O00OOOOO00000O0 :#line:297
        O0O0000OO000O0OOO =f'{KETTLE_HOME}/pan.sh -level=Basic -file={O0O00OOOOO00000O0}{para_str}'#line:298
    else :#line:299
        O0O0000OO000O0OOO =f'{KETTLE_HOME}/kitchen.sh -level=Basic -file={O0O00OOOOO00000O0}{para_str}'#line:300
    print (O0O0000OO000O0OOO )#line:301
    OO00OOOOOOOO0OOOO ,OOO0OOO0000OOOO00 =run_bash (O0O0000OO000O0OOO )#line:305
    if OOO0OOO0000OOOO00 ==0 :#line:306
        print ('{} 完成数据抽取'.format (str (O0O00OOOOO00000O0 )))#line:307
    else :#line:308
        print ('{} 执行错误'.format (O0O00OOOOO00000O0 ))#line:309
        fun_email ('{}/{}出错'.format (OOO0O00OO0OO0O0OO [-2 ],OOO0O00OO0OO0O0OO [-1 ]),str (OO00OOOOOOOO0OOOO ))#line:310
        raise SmartPipError ('Run Kettle Error')#line:311
def hdfsStarrocks (OOO00000O0O0OOO00 ,O000OO000OO00O000 ,para_dict =None ):#line:315
    ""#line:319
    OO0O0O0O0OOOO0OOO =OOO00000O0O0OOO00 .split ('/')#line:320
    print ('strocks load job start')#line:321
    O0O0OO0O00000OOO0 ,O0O00OOOO0O00OO0O =readSqoopFile (OOO00000O0O0OOO00 ,para_dict =para_dict )#line:322
    O0O0OO0O00000OOO0 =O0O0OO0O00000OOO0 .split ('\n')#line:323
    O0OO000000O000OOO ={}#line:324
    O0OO000000O000OOO ['LABEL']=f'{OO0O0O0O0OOOO0OOO[-2]}{OO0O0O0O0OOOO0OOO[-1][:-4]}{int(time.time())}'#line:325
    O0OO000000O000OOO ['HDFS']=HIVE_HOME #line:326
    for O00O0OOO0OO0OOO0O in O0O0OO0O00000OOO0 :#line:327
        OO0OO00O0OO0OOO00 =O00O0OOO0OO0OOO0O .find ('=')#line:328
        if OO0OO00O0OO0OOO00 >0 :#line:329
            O0OO000000O000OOO [O00O0OOO0OO0OOO0O [:OO0OO00O0OO0OOO00 ].strip ()]=O00O0OOO0OO0OOO0O [OO0OO00O0OO0OOO00 +1 :].strip ()#line:330
    O00OO00OO0OO0OO0O =O0OO000000O000OOO .get ('sleepTime')#line:332
    if O00OO00OO0OO0OO0O :#line:333
        O00OO00OO0OO0OO0O =int (O00OO00OO0OO0OO0O )#line:334
        if O00OO00OO0OO0OO0O <30 :#line:335
            O00OO00OO0OO0OO0O =30 #line:336
    else :#line:337
        O00OO00OO0OO0OO0O =30 #line:338
    O000O00OOO000O00O =O0OO000000O000OOO .get ('maxTime')#line:340
    if O000O00OOO000O00O :#line:341
        O000O00OOO000O00O =int (O000O00OOO000O00O )#line:342
        if O000O00OOO000O00O >3600 :#line:343
            O000O00OOO000O00O =3600 #line:344
    else :#line:345
        O000O00OOO000O00O =600 #line:346
    _O0O00000OO000OOO0 (O0O00OOOO0O00OO0O ,db_connect ='mysql',dev ='starrocks',para_dict =O0OO000000O000OOO )#line:348
    time .sleep (O00OO00OO0OO0OO0O )#line:349
    OO0O0O000OO0O0O0O =f'''show load from {O0OO000000O000OOO.get('targetDB')} where label = '{O0OO000000O000OOO['LABEL']}' order by CreateTime desc limit 1 '''#line:350
    OO0O0O0O0000O0000 ='start to check label'#line:351
    try :#line:352
        while True :#line:353
            OO0O0O0O0000O0000 =_O0O00000OO000OOO0 ([OO0O0O000OO0O0O0O ],O000OO000OO00O000 ,db_connect ='mysql',dev ='starrocks')#line:354
            print (OO0O0O0O0000O0000 )#line:355
            O000000OO00OO0OO0 =OO0O0O0O0000O0000 [1 ][2 ]#line:356
            if O000000OO00OO0OO0 =='CANCELLED':#line:357
                raise Exception (f'Starrocks:{O000000OO00OO0OO0}')#line:358
            elif O000000OO00OO0OO0 =='FINISHED':#line:359
                print ('Load completed')#line:360
                break #line:361
            if O000O00OOO000O00O <=0 :#line:362
                raise Exception ('超时未完成')#line:363
            else :#line:364
                time .sleep (O00OO00OO0OO0OO0O )#line:365
                O000O00OOO000O00O =O000O00OOO000O00O -O00OO00OO0OO0OO0O #line:366
    except Exception as OO0000OOO00O0O0OO :#line:367
        print ('{} 执行错误'.format (OOO00000O0O0OOO00 ))#line:368
        fun_email ('{}/{}执行出错'.format (OO0O0O0O0OOOO0OOO [-2 ],OO0O0O0O0OOOO0OOO [-1 ]),str (OO0O0O0O0000O0000 ))#line:369
        raise SmartPipError (str (OO0000OOO00O0O0OO .args ))#line:370
def kafkaStarrocks (OO00OOOOOOO000000 ,O00O00O00OO0OO0O0 ,OOO000OOOOO0000O0 ,OOO00O0O00O0OOO0O ,O00OOO0OO00OOO00O ,dev =''):#line:372
    with open (OO00OOOOOOO000000 ,'r',encoding ='utf8')as O0OO0O000OOO000OO :#line:373
        OOO0O000OOOO0000O =readSqlstr (O0OO0O000OOO000OO .read ().strip (),para_dict =O00OOO0OO00OOO00O )#line:374
    OOO0O000OOOO0000O =OOO0O000OOOO0000O .split ('##')#line:375
    O0O00OOO0OOOO0O00 ={}#line:376
    for OO0OOOO0O00000000 in OOO0O000OOOO0000O :#line:377
        O0000OO00O00OOOO0 =OO0OOOO0O00000000 .find ('=')#line:378
        if O0000OO00O00OOOO0 >0 :#line:379
            O0O00OOO0OO000OOO =OO0OOOO0O00000000 [O0000OO00O00OOOO0 +1 :].replace ('\n',' ').strip ()#line:380
            if O0O00OOO0OO000OOO :#line:381
                O0O00OOO0OOOO0O00 [OO0OOOO0O00000000 [:O0000OO00O00OOOO0 ].strip ()]=O0O00OOO0OO000OOO #line:382
    O0O0OO0000O000000 =O0O00OOO0OOOO0O00 .pop ('topic')#line:383
    O000O0O000OO0OOO0 =O0O00OOO0OOOO0O00 .pop ('table')#line:384
    OOO0OOOOOOOO00O00 =O0O00OOO0OOOO0O00 .keys ()#line:385
    if 'skipError'in OOO0OOOOOOOO00O00 :#line:386
        skipError =O0O00OOO0OOOO0O00 .pop ('skipError')#line:387
    else :#line:388
        skipError =None #line:389
    if 'kafkaConn'in OOO0OOOOOOOO00O00 :#line:390
        OO0OO0O0OO0000OO0 =O0O00OOO0OOOO0O00 .pop ('kafkaConn')#line:391
    else :#line:392
        OO0OO0O0OO0000OO0 ='default'#line:393
    if 'json_root'in OOO0OOOOOOOO00O00 :#line:394
        O0O0O0O0O00O000O0 =O0O00OOO0OOOO0O00 .pop ('json_root')#line:395
    else :#line:396
        O0O0O0O0O00O000O0 =None #line:397
    if 'jsonpaths'in OOO0OOOOOOOO00O00 :#line:398
        O0O0OO0O0000O0O0O =O0O00OOO0OOOO0O00 .get ('jsonpaths')#line:399
        if not O0O0OO0O0000O0O0O .startswith ('['):#line:400
            O0O0OO0O0000O0O0O =O0O0OO0O0000O0O0O .split (',')#line:401
            O0O0OO0O0000O0O0O =json .dumps (['$.'+O0OO0OO0000O0000O .strip ()for O0OO0OO0000O0000O in O0O0OO0O0000O0O0O ])#line:402
            O0O00OOO0OOOO0O00 ['jsonpaths']=O0O0OO0O0000O0O0O #line:403
    OO0000O00OO0OO000 =_OO0OOOO0OO0O0O000 (O0O0OO0000O000000 ,O00O00O00OO0OO0O0 [OO0OO0O0OO0000OO0 ],OOO00O0O00O0OOO0O )#line:404
    def O00OO0000O0O00000 (O0O0OOOOOOO0OO0O0 ):#line:405
        O0O0O00OO000000OO =b''#line:406
        OOOO00O0O0O0000O0 =None #line:407
        if 'format'in OOO0OOOOOOOO00O00 :#line:408
            for OOOO00O0O0O0000O0 in O0O0OOOOOOO0OO0O0 :#line:409
                OOO00O0O000O0O0OO =OOOO00O0O0O0000O0 .value #line:410
                if O0O0O0O0O00O000O0 :#line:411
                    OOO00O0O000O0O0OO =json .loads (OOO00O0O000O0O0OO .decode ('utf8'))#line:412
                    OOO00O0O000O0O0OO =json .dumps (OOO00O0O000O0O0OO [O0O0O0O0O00O000O0 ]).encode ('utf8')#line:413
                if OOO00O0O000O0O0OO .startswith (b'['):#line:414
                    O0O0O00OO000000OO =O0O0O00OO000000OO +b','+OOO00O0O000O0O0OO [1 :-1 ]#line:415
                else :#line:416
                    O0O0O00OO000000OO =O0O0O00OO000000OO +b','+OOO00O0O000O0O0OO #line:417
                if len (O0O0O00OO000000OO )>94857600 :#line:418
                    streamStarrocks (O000O0O000OO0OOO0 ,OOO000OOOOO0000O0 ,O0O00OOO0OOOO0O00 ,O0O0O00OO000000OO ,skipError )#line:419
                    OO0000O00OO0OO000 .write_offset (OOOO00O0O0O0000O0 .partition ,OOOO00O0O0O0000O0 .offset +1 )#line:420
                    O0O0O00OO000000OO =b''#line:421
        else :#line:422
            for OOOO00O0O0O0000O0 in O0O0OOOOOOO0OO0O0 :#line:423
                OOO00O0O000O0O0OO =OOOO00O0O0O0000O0 .value #line:424
                if O0O0O0O0O00O000O0 :#line:425
                    OOO00O0O000O0O0OO =json .loads (OOO00O0O000O0O0OO .decode ('utf8'))#line:426
                    OOO00O0O000O0O0OO =json .dumps (OOO00O0O000O0O0OO [O0O0O0O0O00O000O0 ]).encode ('utf8')#line:427
                O0O0O00OO000000OO =O0O0O00OO000000OO +b'\n'+OOO00O0O000O0O0OO #line:428
                if len (O0O0O00OO000000OO )>94857600 :#line:429
                    streamStarrocks (O000O0O000OO0OOO0 ,OOO000OOOOO0000O0 ,O0O00OOO0OOOO0O00 ,O0O0O00OO000000OO ,skipError )#line:430
                    OO0000O00OO0OO000 .write_offset (OOOO00O0O0O0000O0 .partition ,OOOO00O0O0O0000O0 .offset +1 )#line:431
                    O0O0O00OO000000OO =b''#line:432
        print (O0O0O00OO000000OO [1 :1000 ])#line:433
        if O0O0O00OO000000OO :#line:434
            streamStarrocks (O000O0O000OO0OOO0 ,OOO000OOOOO0000O0 ,O0O00OOO0OOOO0O00 ,O0O0O00OO000000OO ,skipError )#line:435
        return OOOO00O0O0O0000O0 #line:436
    OO0000O00OO0OO000 .consumer_topic (O00OO0000O0O00000 )#line:438
def streamStarrocks (O0O00OOO0O00O0OOO ,OO0OO000OOO0OO0O0 ,OOOO00O000OO0O000 ,OOO0OO0O00OOO0000 ,skipError =False ):#line:440
    import base64 ,uuid #line:441
    OO00000OO0000O00O ,O0O00OOO0O00O0OOO =O0O00OOO0O00O0OOO .split ('.')#line:442
    OO00OO0OO0OO000OO =str (base64 .b64encode (f'{OO0OO000OOO0OO0O0["user"]}:{OO0OO000OOO0OO0O0["password"]}'.encode ('utf-8')),'utf-8')#line:443
    OOO0OO0O00OOO0000 =OOO0OO0O00OOO0000 .strip ()#line:444
    if OOO0OO0O00OOO0000 .startswith (b','):#line:445
        OOOO00O000OO0O000 ['strip_outer_array']='true'#line:446
        OOO0OO0O00OOO0000 =b'['+OOO0OO0O00OOO0000 [1 :]+b']'#line:447
    OOO00O0O00OOO00O0 ={'Content-Type':'application/json','Authorization':f'Basic {OO00OO0OO0OO000OO}','label':f'{O0O00OOO0O00O0OOO}{uuid.uuid4()}',**OOOO00O000OO0O000 }#line:453
    OOO00OO00OOO00O0O =f"{OO0OO000OOO0OO0O0['url']}/api/{OO00000OO0000O00O}/{O0O00OOO0O00O0OOO}/_stream_load"#line:454
    print ('start loading to starrocks....')#line:455
    O0O000000OOOO00OO =requests .put (OOO00OO00OOO00O0O ,headers =OOO00O0O00OOO00O0 ,data =OOO0OO0O00OOO0000 ).json ()#line:456
    print (O0O000000OOOO00OO )#line:457
    if O0O000000OOOO00OO ['Status']=='Fail':#line:458
        if skipError :#line:459
            print (f'Starrocks Load Error, Skip this offset')#line:460
        else :#line:461
            raise Exception ('Starrocks Load Error')#line:462
def point_test (OO0OO0O00000O00O0 ,O00OO0OO0O0OO00O0 ,sleeptime ='',maxtime =''):#line:468
    ""#line:475
    import pymysql #line:476
    if sleeptime :#line:477
        sleeptime =int (sleeptime )#line:478
        sleeptime =sleeptime if sleeptime >60 else 60 #line:479
    if maxtime :#line:480
        maxtime =int (maxtime )#line:481
        maxtime =maxtime if maxtime <60 *60 *2 else 60 *60 *2 #line:482
    else :#line:483
        maxtime =0 #line:484
    O0OO0OO0O000O0OO0 =O00OO0OO0O0OO00O0 ['airflow']#line:485
    OO0OOOOO0OO000OOO =pymysql .connect (user =O0OO0OO0O000O0OO0 ['user'],password =O0OO0OO0O000O0OO0 ['password'],host =O0OO0OO0O000O0OO0 ['host'],port =O0OO0OO0O000O0OO0 ['port'],database =O0OO0OO0O000O0OO0 ['db'],autocommit =True )#line:493
    try :#line:494
        OOO00O000OO0O0000 =OO0OOOOO0OO000OOO .cursor ()#line:495
        OO0OOOOOO00000000 =f"select start_date,state from dag_run where dag_id ='{OO0OO0O00000O00O0}' ORDER BY id desc LIMIT 1"#line:496
        while True :#line:497
            OOO00O000OO0O0000 .execute (OO0OOOOOO00000000 )#line:498
            OOO00O000OOO0O000 =OOO00O000OO0O0000 .fetchall ()#line:499
            if OOO00O000OOO0O000 [0 ][1 ]!='success':#line:500
                if maxtime >0 and OOO00O000OOO0O000 [0 ][1 ]!='failed':#line:501
                    print ('waiting...'+OOO00O000OOO0O000 [0 ][1 ])#line:502
                    time .sleep (sleeptime )#line:503
                    maxtime =maxtime -sleeptime #line:504
                else :#line:505
                    OO0O00O0OO0OO00OO =OOO00O000OOO0O000 [0 ][0 ].strftime ("%Y-%m-%d %H:%M:%S")#line:506
                    O000OO00OOO0O00OO ='所依赖的dag:'+OO0OO0O00000O00O0 +',状态为'+OOO00O000OOO0O000 [0 ][1 ]+'.其最新的执行时间为'+OO0O00O0OO0OO00OO #line:507
                    fun_email (O000OO00OOO0O00OO ,'前置DAG任务未成功')#line:508
                    print (O000OO00OOO0O00OO )#line:509
                    raise SmartPipError ('Run DAG validate Error')#line:510
            else :#line:511
                print ('success...')#line:512
                break #line:513
    except Exception as OOOO000OO0OO0O0OO :#line:514
        print (OOOO000OO0OO0O0OO .args )#line:515
        raise SmartPipError ('DAG validate Error')#line:516
    finally :#line:517
        OO0OOOOO0OO000OOO .close ()#line:518
class connect_db_execute ():#line:523
    def __init__ (O0O0OO000000000OO ,dev =False ,db =''):#line:524
        O0O0OO000000000OO .dev =dev #line:525
    def insert_contents (O00O00OOO000O00O0 ,O0OOO0000OO0OO0O0 ,O0OOOO0O0O000O00O ,per_in =1000 ,connect_dict =None ):#line:527
        OO0O00OO0O0OO00O0 =time .time ()#line:528
        logging .info ('starting to execute insert contents...')#line:529
        if isinstance (connect_dict ,dict ):#line:530
            OO00O00O000OOO0OO =connect_dict ['dbtype']#line:531
        else :#line:532
            if connect_dict =='':#line:533
                OO00O00O000OOO0OO ='oracle'#line:534
            else :#line:535
                OO00O00O000OOO0OO =connect_dict #line:536
            connect_dict =None #line:537
        OO0O00O00O0O0OO00 =getattr (O00O00OOO000O00O0 ,'insert_contents_'+OO00O00O000OOO0OO )#line:538
        OO00OO0OOO00O000O =OO0O00O00O0O0OO00 (O0OOO0000OO0OO0O0 ,O0OOOO0O0O000O00O ,per_in ,connect_dict )#line:539
        logging .info ('execute insert contents time : {}ms'.format (time .time ()-OO0O00OO0O0OO00O0 ))#line:540
        return OO00OO0OOO00O000O #line:541
    def impala (OOOOO0O0O0O0OO0O0 ,OOO0O0O000O0OO000 ,connect_dict =None ):#line:543
        ""#line:544
        from impala .dbapi import connect as impala #line:545
        O0O0O0OOO00O00O00 =impala (user =connect_dict ['user'],password =connect_dict ['password'],host =connect_dict ['host'],port =int (connect_dict ['port']),auth_mechanism ='PLAIN')#line:552
        OO000000OOOO0O0OO =O0O0O0OOO00O00O00 .cursor ()#line:553
        OOO00O00OOOOO00O0 =r'^insert\s|^update\s|^truncate\s|^delete\s|^load\s|^refresh\s|^upsert\s'#line:554
        OOOOO00000O000OOO =None #line:555
        for O000OOOOOOOOO0OOO in OOO0O0O000O0OO000 :#line:556
            print (O000OOOOOOOOO0OOO )#line:557
            O000OOOOOOOOO0OOO =O000OOOOOOOOO0OOO .strip ()#line:558
            if not O000OOOOOOOOO0OOO :#line:559
                continue #line:560
            if re .search (OOO00O00OOOOO00O0 ,O000OOOOOOOOO0OOO ,re .I |re .IGNORECASE ):#line:561
                OO000000OOOO0O0OO .execute (O000OOOOOOOOO0OOO )#line:562
            else :#line:563
                OO000000OOOO0O0OO .execute (O000OOOOOOOOO0OOO )#line:564
                try :#line:565
                    OOOOO00000O000OOO =OO000000OOOO0O0OO .fetchall ()#line:566
                except Exception as O0O000OOOOOO0O0O0 :#line:567
                    print (O0O000OOOOOO0O0O0 .args )#line:568
        O0O0O0OOO00O00O00 .close ()#line:569
        return OOOOO00000O000OOO #line:570
    def hive (OOOOO0OO00O0OOOO0 ,OOOOOOO0O00OO0O00 ,connect_dict =None ):#line:572
        ""#line:573
        from impala .dbapi import connect as impala #line:574
        OO0OO0O0OO0O0000O =impala (user =connect_dict ['user'],password =connect_dict ['password'],host =connect_dict ['host'],port =int (connect_dict ['port']),auth_mechanism ='PLAIN')#line:581
        OO000O0OO000O00O0 =OO0OO0O0OO0O0000O .cursor ()#line:582
        O0OO000OOOOOOOOOO =r'^insert\s|^update\s|^truncate\s|^delete\s|^load\s'#line:583
        O0OOOO0O0O0OO0OOO =None #line:584
        for OO00OO0O00O00O00O in OOOOOOO0O00OO0O00 :#line:585
            OO00OO0O00O00O00O =OO00OO0O00O00O00O .strip ()#line:586
            if not OO00OO0O00O00O00O :#line:587
                continue #line:588
            print (OO00OO0O00O00O00O )#line:589
            if OO00OO0O00O00O00O .startswith ('refresh'):#line:590
                connect_dict ['port']=21050 #line:591
                OOOOO0OO00O0OOOO0 .impala ([OO00OO0O00O00O00O ],connect_dict =connect_dict )#line:592
            else :#line:593
                if re .search (O0OO000OOOOOOOOOO ,OO00OO0O00O00O00O ,re .I |re .IGNORECASE ):#line:594
                    OO000O0OO000O00O0 .execute (OO00OO0O00O00O00O )#line:595
                else :#line:596
                    OO000O0OO000O00O0 .execute (OO00OO0O00O00O00O )#line:597
                    try :#line:598
                        O0OOOO0O0O0OO0OOO =OO000O0OO000O00O0 .fetchall ()#line:599
                    except Exception as OOO0OO0O0O0000O00 :#line:600
                        print (OOO0OO0O0O0000O00 .args )#line:601
        OO0OO0O0OO0O0000O .close ()#line:602
        return O0OOOO0O0O0OO0OOO #line:603
    def mysql (OOOO0OO00OO0OOOO0 ,OO0O0OO0OOOO00OO0 ,connect_dict =None ):#line:605
        import pymysql #line:606
        OOOO00OO0O0O00OO0 =pymysql .connect (user =connect_dict ['user'],password =connect_dict ['password'],host =connect_dict ['host'],port =connect_dict ['port'],database =connect_dict ['db'])#line:613
        try :#line:614
            OO0O000OO000OOOOO =OOOO00OO0O0O00OO0 .cursor ()#line:615
            OO000O0O00OO00000 =r'^insert\s|^update\s|^truncate\s|^delete\s|^load\s'#line:616
            for O00O00O00O0OOO000 in OO0O0OO0OOOO00OO0 :#line:617
                O00O00O00O0OOO000 =O00O00O00O0OOO000 .strip ()#line:618
                if not O00O00O00O0OOO000 :#line:619
                    continue #line:620
                print (O00O00O00O0OOO000 )#line:621
                if re .search (OO000O0O00OO00000 ,O00O00O00O0OOO000 ,re .I |re .IGNORECASE ):#line:622
                    try :#line:623
                        OO0O000OO000OOOOO .execute (O00O00O00O0OOO000 )#line:624
                        OOOO00OO0O0O00OO0 .commit ()#line:625
                    except Exception as O0OO00O0O0OOOOOO0 :#line:626
                        OOOO00OO0O0O00OO0 .rollback ()#line:627
                        raise O0OO00O0O0OOOOOO0 #line:628
                else :#line:629
                    OO0O000OO000OOOOO .execute (O00O00O00O0OOO000 )#line:630
                    OOO0O0O00O00000O0 =OO0O000OO000OOOOO .fetchall ()#line:631
                    OOO0O0O00O00000O0 =[[O00O00000O0OO000O [0 ]for O00O00000O0OO000O in OO0O000OO000OOOOO .description ]]+list (OOO0O0O00O00000O0 )#line:632
                    return OOO0O0O00O00000O0 #line:633
        except Exception as OO000OO0000OO0OOO :#line:634
            raise OO000OO0000OO0OOO #line:635
        finally :#line:636
            OOOO00OO0O0O00OO0 .close ()#line:637
    def starrocks (OOOO00O0OO0OO0000 ,OO0O0000O000000OO ,connect_dict =None ):#line:639
        return OOOO00O0OO0OO0000 .mysql (OO0O0000O000000OO ,connect_dict )#line:640
    def oracle (O0OOOO0O00OOOO0O0 ,O0O0O00O0O00O0OO0 ,connect_dict =None ):#line:642
        import cx_Oracle #line:643
        OO0OO0O00O00000OO ='{}/{}@{}/{}'.format (connect_dict ['user'],connect_dict ['password'],connect_dict ['host'],connect_dict ['db'])#line:648
        O00000O00OO00OO0O =cx_Oracle .connect (OO0OO0O00O00000OO )#line:649
        try :#line:650
            OO0OOO0OO0O0OOOO0 =O00000O00OO00OO0O .cursor ()#line:651
            OO0OOO00O00O00000 =r'^insert\s|^update\s|^truncate\s|^delete\s|^comment\s'#line:652
            for O0O0OO0O00O000OOO in O0O0O00O0O00O0OO0 :#line:653
                O0O0OO0O00O000OOO =O0O0OO0O00O000OOO .strip ()#line:654
                if not O0O0OO0O00O000OOO :#line:655
                    continue #line:656
                if re .search (OO0OOO00O00O00000 ,O0O0OO0O00O000OOO ,re .I ):#line:657
                    try :#line:658
                        OO0OOO0OO0O0OOOO0 .execute (O0O0OO0O00O000OOO )#line:659
                        O00000O00OO00OO0O .commit ()#line:660
                    except Exception as OOOO0O00OO0OO0OOO :#line:661
                        if O0O0OO0O00O000OOO .startswith ('comment'):#line:662
                            print ('err:',O0O0OO0O00O000OOO )#line:663
                            continue #line:664
                        O00000O00OO00OO0O .rollback ()#line:665
                        raise OOOO0O00OO0OO0OOO #line:666
                else :#line:667
                    OO0OOO0OO0O0OOOO0 .execute (O0O0OO0O00O000OOO )#line:668
                    OO000O000000O0000 =OO0OOO0OO0O0OOOO0 .fetchall ()#line:669
                    OO000O000000O0000 =[[O0O0OO0O00OOOO0OO [0 ]for O0O0OO0O00OOOO0OO in OO0OOO0OO0O0OOOO0 .description ]]+list (OO000O000000O0000 )#line:670
                    return OO000O000000O0000 #line:671
        except Exception as O0OOOOO00O0OO0OOO :#line:672
            raise O0OOOOO00O0OO0OOO #line:673
        finally :#line:674
            O00000O00OO00OO0O .close ()#line:675
    def gp (OOO00OO000OOO0000 ,O0O000OO0OO00OO00 ,connect_dict =None ):#line:677
        import psycopg2 #line:678
        OO0OO00000OOO000O =psycopg2 .connect (user =connect_dict ['user'],password =connect_dict ['password'],host =connect_dict ['host'],port =connect_dict ['port'],database =connect_dict ['db'])#line:685
        try :#line:686
            O00O00O000000OOOO =OO0OO00000OOO000O .cursor ()#line:687
            O000OO000O0O000O0 =r'^insert\s|^update\s|^truncate\s|^delete\s'#line:688
            for O00000OO00OOOOOOO in O0O000OO0OO00OO00 :#line:689
                O00000OO00OOOOOOO =O00000OO00OOOOOOO .strip ()#line:690
                if not O00000OO00OOOOOOO :#line:691
                    continue #line:692
                if re .search (O000OO000O0O000O0 ,O00000OO00OOOOOOO ,re .I |re .IGNORECASE ):#line:693
                    try :#line:694
                        O00O00O000000OOOO .execute (O00000OO00OOOOOOO )#line:695
                        OO0OO00000OOO000O .commit ()#line:696
                    except Exception as OO00000OO0OOOOO00 :#line:697
                        OO0OO00000OOO000O .rollback ()#line:698
                        raise OO00000OO0OOOOO00 #line:699
                else :#line:700
                    O00O00O000000OOOO .execute (O00000OO00OOOOOOO )#line:701
                    O0OO0OOOOOOO00O0O =O00O00O000000OOOO .fetchall ()#line:702
                    O0OO0OOOOOOO00O0O =[[O00OOO0OO0000OO00 [0 ]for O00OOO0OO0000OO00 in O00O00O000000OOOO .description ]]+list (O0OO0OOOOOOO00O0O )#line:703
                    return O0OO0OOOOOOO00O0O #line:704
        except Exception as OO0O00OOO0OOOO0OO :#line:705
            raise OO0O00OOO0OOOO0OO #line:706
        finally :#line:707
            OO0OO00000OOO000O .close ()#line:708
    def execute_sql_list (O0000OO0OOO0OOO00 ,OO0O0OO000OO00OO0 ,db_connect ='',connect_dict =None ):#line:710
        if db_connect =='':db_connect ='oracle'#line:711
        O00OOO00OOO00O000 =getattr (O0000OO0OOO0OOO00 ,db_connect )#line:712
        return O00OOO00OOO00O000 (OO0O0OO000OO00OO0 ,connect_dict =connect_dict )#line:713
    def excute_proc (O0OOOO000O0O0OO0O ,OOO00O0OO00000OOO ,OOOOOO00OO000OOOO ,proc_para =None ):#line:715
        import cx_Oracle #line:716
        OO000O0OOO0OO0OO0 ='{}/{}@{}/{}'.format (OOOOOO00OO000OOOO ['user'],OOOOOO00OO000OOOO ['password'],OOOOOO00OO000OOOO ['host'],OOOOOO00OO000OOOO ['db'])#line:722
        OOOO0O0O0O0OO0OO0 =cx_Oracle .connect (OO000O0OOO0OO0OO0 )#line:723
        try :#line:725
            OOOOO0OO0OOO0OO0O =OOOO0O0O0O0OO0OO0 .cursor ()#line:726
            print ("开始执行过程:{}  参数: {}".format (OOO00O0OO00000OOO ,proc_para ))#line:727
            if proc_para is None :#line:728
                O0O00O0OOO0OOOO0O =OOOOO0OO0OOO0OO0O .callproc (OOO00O0OO00000OOO )#line:729
                OOOO0O0O0O0OO0OO0 .commit ()#line:730
            else :#line:731
                O0O00O0OOO0OOOO0O =OOOOO0OO0OOO0OO0O .callproc (OOO00O0OO00000OOO ,proc_para )#line:733
                OOOO0O0O0O0OO0OO0 .commit ()#line:734
            OOOOO0OO0OOO0OO0O .close ()#line:735
            OOOO0O0O0O0OO0OO0 .close ()#line:736
            print (O0O00O0OOO0OOOO0O )#line:737
        except Exception as O0O0000OOO0OO0O0O :#line:738
            OOOO0O0O0O0OO0OO0 .rollback ()#line:739
            OOOO0O0O0O0OO0OO0 .close ()#line:740
            raise O0O0000OOO0OO0O0O #line:742
        return True #line:743
    def insert_contents_oracle (OO0O00O00OOOOO0O0 ,O000O0O00O000OO0O ,O00OO0O0OO0O0OO00 ,per_in =100 ,connect_dict =None ):#line:745
        import cx_Oracle #line:746
        OOO0OOO000OO0OOOO ='{}/{}@{}:{}/{}'.format (connect_dict ['user'],connect_dict ['password'],connect_dict ['host'],connect_dict ['port'],connect_dict ['db'])#line:752
        OO000O00OO00OOOO0 =cx_Oracle .connect (OOO0OOO000OO0OOOO )#line:753
        O0O0O00O0OOO00O00 =OO000O00OO00OOOO0 .cursor ()#line:754
        OO00OO00OOOO00O00 =' into {} values {}'#line:755
        O000000O000OO0000 =''#line:756
        O0OOOO00OO0O00O0O =len (O000O0O00O000OO0O )#line:757
        logging .info ("total {} records need to insert table {}: ".format (O0OOOO00OO0O00O0O ,O00OO0O0OO0O0OO00 ))#line:758
        try :#line:759
            for O0O000000OOOO0O0O in range (O0OOOO00OO0O00O0O ):#line:760
                O000000O000OO0000 =O000000O000OO0000 +OO00OO00OOOO00O00 .format (O00OO0O0OO0O0OO00 ,tuple (O000O0O00O000OO0O [O0O000000OOOO0O0O ]))+'\n'#line:761
                if (O0O000000OOOO0O0O +1 )%per_in ==0 or O0O000000OOOO0O0O ==O0OOOO00OO0O00O0O -1 :#line:762
                    OO0OOOOO0O0O0O0OO ='insert all '+O000000O000OO0000 +'select 1 from dual'#line:763
                    logging .debug (OO0OOOOO0O0O0O0OO )#line:764
                    O0O0O00O0OOO00O00 .execute (OO0OOOOO0O0O0O0OO )#line:765
                    OO000O00OO00OOOO0 .commit ()#line:766
                    O000000O000OO0000 =''#line:767
            return str (O0OOOO00OO0O00O0O )#line:768
        except Exception as OOOOOO0OOO0OOO00O :#line:769
            try :#line:770
                OO000O00OO00OOOO0 .rollback ()#line:771
                O0O0O00O0OOO00O00 .execute ("delete from {} where UPLOADTIME = '{}'".format (O00OO0O0OO0O0OO00 ,O000O0O00O000OO0O [0 ][-1 ]))#line:772
                OO000O00OO00OOOO0 .commit ()#line:773
            except Exception :#line:774
                logging .error ('can not delete by uploadtime')#line:775
            finally :#line:776
                raise OOOOOO0OOO0OOO00O #line:777
        finally :#line:778
            OO000O00OO00OOOO0 .close ()#line:779
class _OO0OOOO0OO0O0O000 (object ):#line:782
    connect =None #line:783
    def __init__ (O0O0O0000O0O0OOOO ,O00O000000O00O00O ,OOOOOOOO00O00O0OO ,OO00OO0O0000OO0OO ):#line:784
        O0O0O0000O0O0OOOO .offset =None #line:785
        O0O0O0000O0O0OOOO .db_err_count =0 #line:786
        O0O0O0000O0O0OOOO .topic =O00O000000O00O00O #line:787
        O0O0O0000O0O0OOOO .kafkaconfig =OOOOOOOO00O00O0OO #line:788
        O0O0O0000O0O0OOOO .offsetDict ={}#line:789
        O0O0O0000O0O0OOOO .current_dir =OO00OO0O0000OO0OO #line:790
        try :#line:791
            O0O0O0000O0O0OOOO .consumer =O0O0O0000O0O0OOOO .connect_kafka_customer ()#line:792
        except Exception as O0OOOO00OO0OOOO0O :#line:793
            O0OOOO00OO0OOOO0O ='kafka无法连接','ErrLocation：{}\n'.format (O00O000000O00O00O )+str (O0OOOO00OO0OOOO0O )+',kafka消费者无法创建'#line:794
            raise O0OOOO00OO0OOOO0O #line:795
    def get_toggle_or_offset (OO0OO0OOOOOO000OO ,O0O00O0O0OOOO000O ,OOO0000O0O00O00O0 ):#line:798
        ""#line:799
        O0OO0OO0OOOOO0000 =0 #line:800
        try :#line:801
            O000O00OO00O000O0 =f"{OO0OO0OOOOOO000OO.current_dir}/kafka/{O0O00O0O0OOOO000O}_offset_{OOO0000O0O00O00O0}.txt"#line:802
            if os .path .exists (O000O00OO00O000O0 ):#line:803
                OO0OOO00000OOO00O =open (O000O00OO00O000O0 ).read ()#line:804
                if OO0OOO00000OOO00O :#line:805
                    O0OO0OO0OOOOO0000 =int (OO0OOO00000OOO00O )#line:806
            else :#line:807
                with open (O000O00OO00O000O0 ,encoding ='utf-8',mode ='a')as OO0000OO0O000OO00 :#line:808
                    OO0000OO0O000OO00 .close ()#line:809
        except Exception as OO00OOO0O0OOOOOO0 :#line:810
            print (f"读取失败：{OO00OOO0O0OOOOOO0}")#line:811
            raise OO00OOO0O0OOOOOO0 #line:812
        return O0OO0OO0OOOOO0000 #line:813
    def write_offset (O0O0O0O00OOOOOOO0 ,O0O0O0OO0OOOOOO00 ,offset =None ):#line:815
        ""#line:818
        if O0O0O0O00OOOOOOO0 .topic and offset :#line:819
            OO0O00OOOOOO0OOO0 =f"{O0O0O0O00OOOOOOO0.current_dir}/kafka/{O0O0O0O00OOOOOOO0.topic}_offset_{O0O0O0OO0OOOOOO00}.txt"#line:821
            try :#line:822
                with open (OO0O00OOOOOO0OOO0 ,'w')as O0000OOO0OOOOOOOO :#line:823
                    O0000OOO0OOOOOOOO .write (str (offset ))#line:824
                    O0000OOO0OOOOOOOO .close ()#line:825
            except Exception as OOO0O0OO00OOO0O00 :#line:826
                print (f"写入offset出错：{OOO0O0OO00OOO0O00}")#line:827
                raise OOO0O0OO00OOO0O00 #line:828
    def connect_kafka_customer (O0O0OOO00O0O0O000 ):#line:830
        ""#line:831
        OOOOOO0OO000000O0 =KafkaConsumer (**O0O0OOO00O0O0O000 .kafkaconfig )#line:832
        return OOOOOO0OO000000O0 #line:833
    def parse_data (O000000000O0O0OOO ,O00OOOOO0OOOO0000 ):#line:835
        ""#line:840
        return dict ()#line:841
    def gen_sql (OO000O0OO00OO000O ,OOOOO0OOOOO00O0O0 ):#line:843
        ""#line:848
        O0O0O000OOOO000OO =[]#line:849
        for O00O0OO0OOO00O0O0 in OOOOO0OOOOO00O0O0 :#line:850
            O0O0O000OOOO000OO .append (str (tuple (O00O0OO0OOO00O0O0 )))#line:852
        return ','.join (O0O0O000OOOO000OO )#line:853
    def dispose_kafka_data (O0OO0O0000OO00OOO ,OOO0O00O00OO00O00 ):#line:855
        ""#line:860
        pass #line:861
    def get_now_time (OO0O00000OO0O00OO ):#line:863
        ""#line:867
        O0OO0OOO0OO0O0000 =int (time .time ())#line:868
        return time .strftime ('%Y-%m-%d %H:%M:%S',time .localtime (O0OO0OOO0OO0O0000 ))#line:869
    def tran_data (O000OO00OOO000000 ,O00O00000OOO0O0O0 ,OO00O0O000O0000OO ):#line:871
        ""#line:877
        O000O0OOO0O0O0000 =O00O00000OOO0O0O0 .get (OO00O0O000O0000OO ,"")#line:878
        O000O0OOO0O0O0000 =""if O000O0OOO0O0O0000 is None else O000O0OOO0O0O0000 #line:879
        return str (O000O0OOO0O0O0000 )#line:880
    def consumer_data (O0O0OO000O0OO00OO ,O0000OOO0OOO00OOO ,O000O00O0OO0OO000 ,OO0O0O000000OO000 ):#line:882
        ""#line:889
        if O0O0OO000O0OO00OO .consumer :#line:890
            O0O0OO000O0OO00OO .consumer .assign ([TopicPartition (topic =O0O0OO000O0OO00OO .topic ,partition =O0000OOO0OOO00OOO )])#line:891
            OO00O0OOOO00O0000 =TopicPartition (topic =O0O0OO000O0OO00OO .topic ,partition =O0000OOO0OOO00OOO )#line:893
            OOOOOOOO0O00OO0OO =O0O0OO000O0OO00OO .consumer .beginning_offsets ([OO00O0OOOO00O0000 ])#line:894
            O0OOO000O00OOOO0O =OOOOOOOO0O00OO0OO .get (OO00O0OOOO00O0000 )#line:895
            OOOOO0O00O0O0OOOO =O0O0OO000O0OO00OO .consumer .end_offsets ([OO00O0OOOO00O0000 ])#line:896
            O0000O000O0OO0O0O =OOOOO0O00O0O0OOOO .get (OO00O0OOOO00O0000 )#line:897
            print (f'建立消费者, {O0000OOO0OOO00OOO}分区, 最小offset:{O0OOO000O00OOOO0O}, 最大offset:{O0000O000O0OO0O0O}')#line:898
            if O000O00O0OO0OO000 >=O0000O000O0OO0O0O :#line:899
                print (f'消费offset：{O000O00O0OO0OO000} 大于最大offset:{O0000O000O0OO0O0O}, 本次不消费')#line:900
                return #line:901
            if O000O00O0OO0OO000 <O0OOO000O00OOOO0O :#line:902
                print (f'Warning: 消费offset：{O000O00O0OO0OO000} 小于最小offset:{O0OOO000O00OOOO0O}')#line:903
                O000O00O0OO0OO000 =O0OOO000O00OOOO0O #line:904
            O0O0OO000O0OO00OO .consumer .seek (TopicPartition (topic =O0O0OO000O0OO00OO .topic ,partition =O0000OOO0OOO00OOO ),offset =O000O00O0OO0OO000 )#line:905
            print (f'消费{O0000OOO0OOO00OOO}分区, 开始消费offset：{O000O00O0OO0OO000}')#line:906
            OO0O0O000O00OO0OO =OO0O0O000000OO000 (O0O0OO000O0OO00OO .consumer )#line:907
            O000O00O0OO0OO000 =OO0O0O000O00OO0OO.offset +1 #line:908
            O0O0OO000O0OO00OO .offsetDict [O0000OOO0OOO00OOO ]=O000O00O0OO0OO000 #line:911
            O0O0OO000O0OO00OO .write_offset (O0000OOO0OOO00OOO ,O000O00O0OO0OO000 )#line:912
    def consumer_topic (OO00O0O000OOOO00O ,OO0OO0O00OO00O000 ):#line:914
        print (f"topic: {OO00O0O000OOOO00O.topic}")#line:915
        print ('开始解析。')#line:916
        O0O0O0O000OOO000O =OO00O0O000OOOO00O .consumer .partitions_for_topic (topic =OO00O0O000OOOO00O .topic )#line:918
        for O0O0OOOOO00OO0O00 in O0O0O0O000OOO000O :#line:919
            O0OOO0O0O0OO0O000 =OO00O0O000OOOO00O .get_toggle_or_offset (OO00O0O000OOOO00O .topic ,O0O0OOOOO00OO0O00 )#line:920
            O00O0O00O0OOOOOO0 =None if O0OOO0O0O0OO0O000 <0 else O0OOO0O0O0OO0O000 #line:922
            OO00O0O000OOOO00O .consumer_data (O0O0OOOOO00OO0O00 ,O00O0O00O0OOOOOO0 ,OO0OO0O00OO00O000 )#line:923
    def save_all_offset (OO00O000O00OO0O00 ):#line:925
        for OO000O00000OO00OO ,O0O000OO0OOOO0OO0 in OO00O000O00OO0O00 .offsetDict :#line:926
            OO00O000O00OO0O00 .write_offset (OO000O00000OO00OO ,O0O000OO0OOOO0OO0 )#line:927



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Base task runner"""
import os
import subprocess
import threading
from tempfile import NamedTemporaryFile
from typing import Optional, Union

from airflow.configuration import conf
from airflow.exceptions import AirflowConfigException
from airflow.models.taskinstance import load_error_file
from airflow.utils.configuration import tmp_configuration_copy
from airflow.utils.log.logging_mixin import LoggingMixin
from airflow.utils.net import get_hostname
from airflow.utils.platform import getuser

PYTHONPATH_VAR = 'PYTHONPATH'


class BaseTaskRunner(LoggingMixin):
    """
    Runs Airflow task instances by invoking the `airflow tasks run` command with raw
    mode enabled in a subprocess.

    :param local_task_job: The local task job associated with running the
        associated task instance.
    :type local_task_job: airflow.jobs.local_task_job.LocalTaskJob
    """

    def __init__(self, local_task_job):
        # Pass task instance context into log handlers to setup the logger.
        super().__init__(local_task_job.task_instance)
        self._task_instance = local_task_job.task_instance

        popen_prepend = []
        if self._task_instance.run_as_user:
            self.run_as_user = self._task_instance.run_as_user
        else:
            try:
                self.run_as_user = conf.get('core', 'default_impersonation')
            except AirflowConfigException:
                self.run_as_user = None

        # Add sudo commands to change user if we need to. Needed to handle SubDagOperator
        # case using a SequentialExecutor.
        self.log.debug("Planning to run as the %s user", self.run_as_user)
        if self.run_as_user and (self.run_as_user != getuser()):
            # We want to include any environment variables now, as we won't
            # want to have to specify them in the sudo call - they would show
            # up in `ps` that way! And run commands now, as the other user
            # might not be able to run the cmds to get credentials
            cfg_path = tmp_configuration_copy(chmod=0o600)

            # Give ownership of file to user; only they can read and write
            subprocess.call(['sudo', 'chown', self.run_as_user, cfg_path], close_fds=True)

            # propagate PYTHONPATH environment variable
            pythonpath_value = os.environ.get(PYTHONPATH_VAR, '')
            popen_prepend = ['sudo', '-E', '-H', '-u', self.run_as_user]

            if pythonpath_value:
                popen_prepend.append(f'{PYTHONPATH_VAR}={pythonpath_value}')

        else:
            # Always provide a copy of the configuration file settings. Since
            # we are running as the same user, and can pass through environment
            # variables then we don't need to include those in the config copy
            # - the runner can read/execute those values as it needs
            cfg_path = tmp_configuration_copy(chmod=0o600)

        self._error_file = NamedTemporaryFile(delete=True)
        self._cfg_path = cfg_path
        self._command = (
            popen_prepend
            + self._task_instance.command_as_list(
                raw=True,
                pickle_id=local_task_job.pickle_id,
                mark_success=local_task_job.mark_success,
                job_id=local_task_job.id,
                pool=local_task_job.pool,
                cfg_path=cfg_path,
            )
            + ["--error-file", self._error_file.name]
        )
        self.process = None

    def deserialize_run_error(self) -> Optional[Union[str, Exception]]:
        """Return task runtime error if its written to provided error file."""
        return load_error_file(self._error_file)

    def _read_task_logs(self, stream):
        while True:
            line = stream.readline()
            if isinstance(line, bytes):
                line = line.decode('utf-8')
            if not line:
                break
            self.log.info(
                'Job %s: Subtask %s %s',
                self._task_instance.job_id,
                self._task_instance.task_id,
                line.rstrip('\n'),
            )

    def run_command(self, run_with=None):
        """
        Run the task command.

        :param run_with: list of tokens to run the task command with e.g. ``['bash', '-c']``
        :type run_with: list
        :return: the process that was run
        :rtype: subprocess.Popen
        """
        run_with = run_with or []
        full_cmd = run_with + self._command

        self.log.info("Running on host: %s", get_hostname())
        self.log.info('Running: %s', full_cmd)

        proc = subprocess.Popen(
            full_cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
            close_fds=True,
            env=os.environ.copy(),
            preexec_fn=os.setsid,
        )

        # Start daemon thread to read subprocess logging output
        log_reader = threading.Thread(
            target=self._read_task_logs,
            args=(proc.stdout,),
        )
        log_reader.daemon = True
        log_reader.start()
        return proc

    def start(self):
        """Start running the task instance in a subprocess."""
        raise NotImplementedError()

    def return_code(self) -> Optional[int]:
        """
        :return: The return code associated with running the task instance or
            None if the task is not yet done.
        :rtype: int
        """
        raise NotImplementedError()

    def terminate(self) -> None:
        """Force kill the running task instance."""
        raise NotImplementedError()

    def on_finish(self) -> None:
        """A callback that should be called when this is done running."""
        if self._cfg_path and os.path.isfile(self._cfg_path):
            if self.run_as_user:
                subprocess.call(['sudo', 'rm', self._cfg_path], close_fds=True)
            else:
                os.remove(self._cfg_path)
        self._error_file.close()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Setup.py for the Airflow project."""
import glob
import logging
import os
import subprocess
import sys
import unittest
from copy import deepcopy
from distutils import log
from os.path import dirname, relpath
from textwrap import wrap
from typing import Dict, List

from setuptools import Command, Distribution, find_namespace_packages, setup
from setuptools.command.develop import develop as develop_orig
from setuptools.command.install import install as install_orig

# Controls whether providers are installed from packages or directly from sources
# It is turned on by default in case of development environments such as Breeze
# And it is particularly useful when you add a new provider and there is no
# PyPI version to install the provider package from
INSTALL_PROVIDERS_FROM_SOURCES = 'INSTALL_PROVIDERS_FROM_SOURCES'
PY39 = sys.version_info >= (3, 9)

logger = logging.getLogger(__name__)

version = '2.1.4'

my_dir = dirname(__file__)


def airflow_test_suite() -> unittest.TestSuite:
    """Test suite for Airflow tests"""
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover(os.path.join(my_dir, 'tests'), pattern='test_*.py')
    return test_suite


class CleanCommand(Command):
    """
    Command to tidy up the project root.
    Registered as cmdclass in setup() so it can be called with ``python setup.py extra_clean``.
    """

    description = "Tidy up the project root"
    user_options: List[str] = []

    def initialize_options(self) -> None:
        """Set default values for options."""

    def finalize_options(self) -> None:
        """Set final values for options."""

    @staticmethod
    def rm_all_files(files: List[str]) -> None:
        """Remove all files from the list"""
        for file in files:
            try:
                os.remove(file)
            except Exception as e:
                logger.warning("Error when removing %s: %s", file, e)

    def run(self) -> None:
        """Remove temporary files and directories."""
        os.chdir(my_dir)
        self.rm_all_files(glob.glob('./build/*'))
        self.rm_all_files(glob.glob('./**/__pycache__/*', recursive=True))
        self.rm_all_files(glob.glob('./**/*.pyc', recursive=True))
        self.rm_all_files(glob.glob('./dist/*'))
        self.rm_all_files(glob.glob('./*.egg-info'))
        self.rm_all_files(glob.glob('./docker-context-files/*.whl'))
        self.rm_all_files(glob.glob('./docker-context-files/*.tgz'))


class CompileAssets(Command):
    """
    Compile and build the frontend assets using yarn and webpack.
    Registered as cmdclass in setup() so it can be called with ``python setup.py compile_assets``.
    """

    description = "Compile and build the frontend assets"
    user_options: List[str] = []

    def initialize_options(self) -> None:
        """Set default values for options."""

    def finalize_options(self) -> None:
        """Set final values for options."""

    def run(self) -> None:
        """Run a command to compile and build assets."""
        subprocess.check_call('./airflow/www/compile_assets.sh')


class ListExtras(Command):
    """
    List all available extras
    Registered as cmdclass in setup() so it can be called with ``python setup.py list_extras``.
    """

    description = "List available extras"
    user_options: List[str] = []

    def initialize_options(self) -> None:
        """Set default values for options."""

    def finalize_options(self) -> None:
        """Set final values for options."""

    def run(self) -> None:
        """List extras."""
        print("\n".join(wrap(", ".join(EXTRAS_REQUIREMENTS.keys()), 100)))


def git_version(version_: str) -> str:
    """
    Return a version to identify the state of the underlying git repo. The version will
    indicate whether the head of the current git-backed working directory is tied to a
    release tag or not : it will indicate the former with a 'release:{version}' prefix
    and the latter with a '.dev0' suffix. Following the prefix will be a sha of the current
    branch head. Finally, a "dirty" suffix is appended to indicate that uncommitted
    changes are present.

    :param str version_: Semver version
    :return: Found Airflow version in Git repo
    :rtype: str
    """
    try:
        import git

        try:
            repo = git.Repo(os.path.join(*[my_dir, '.git']))
        except git.NoSuchPathError:
            logger.warning('.git directory not found: Cannot compute the git version')
            return ''
        except git.InvalidGitRepositoryError:
            logger.warning('Invalid .git directory not found: Cannot compute the git version')
            return ''
    except ImportError:
        logger.warning('gitpython not found: Cannot compute the git version.')
        return ''
    if repo:
        sha = repo.head.commit.hexsha
        if repo.is_dirty():
            return f'.dev0+{sha}.dirty'
        # commit is clean
        return f'.release:{version_}+{sha}'
    return 'no_git_version'


def write_version(filename: str = os.path.join(*[my_dir, "airflow", "git_version"])) -> None:
    """
    Write the Semver version + git hash to file, e.g. ".dev0+2f635dc265e78db6708f59f68e8009abb92c1e65".

    :param str filename: Destination file to write
    """
    text = f"{git_version(version)}"
    with open(filename, 'w') as file:
        file.write(text)


# 'Start dependencies group' and 'Start dependencies group' are mark for ./scripts/ci/check_order_setup.py
# If you change this mark you should also change ./scripts/ci/check_order_setup.py
# Start dependencies group
amazon = [
    'boto3>=1.15.0,<1.18.0',
    'watchtower~=1.0.6',
    'jsonpath_ng>=1.5.3',
]
apache_beam = [
    'apache-beam>=2.20.0',
]
asana = ['asana>=0.10', 'cached-property>=1.5.2']
async_packages = [
    'eventlet>= 0.9.7',
    'gevent>=0.13',
    'greenlet>=0.4.9',
]
atlas = [
    'atlasclient>=0.1.2',
]
azure = [
    'azure-batch>=8.0.0',
    'azure-cosmos>=3.0.1,<4',
    'azure-datalake-store>=0.0.45',
    'azure-identity>=1.3.1',
    'azure-keyvault>=4.1.0',
    'azure-kusto-data>=0.0.43,<0.1',
    'azure-mgmt-containerinstance>=1.5.0,<2.0',
    'azure-mgmt-datafactory>=1.0.0,<2.0',
    'azure-mgmt-datalake-store>=0.5.0',
    'azure-mgmt-resource>=2.2.0',
    'azure-storage-blob>=12.7.0',
    'azure-storage-common>=2.1.0',
    'azure-storage-file>=2.1.0',
]
cassandra = [
    'cassandra-driver>=3.13.0,<4',
]
celery = [
    'celery~=4.4.2',
    'flower>=0.7.3, <1.0',
    'vine~=1.3',  # https://stackoverflow.com/questions/32757259/celery-no-module-named-five
]
cgroups = [
    'cgroupspy>=0.1.4',
]
cloudant = [
    'cloudant>=2.0',
]
dask = [
    'cloudpickle>=1.4.1, <1.5.0',
    'dask<2021.3.1;python_version<"3.7"',  # dask stopped supporting python 3.6 in 2021.3.1 version
    'dask>=2.9.0, <2021.6.1;python_version>="3.7"',  # dask 2021.6.1 does not work with `distributed`
    'distributed>=2.11.1, <2.20',
]
databricks = [
    'requests>=2.26.0, <3',
]
datadog = [
    'datadog>=0.14.0',
]
deprecated_api = [
    'requests>=2.26.0',
]
doc = [
    # Sphinx is limited to < 3.5.0 because of https://github.com/sphinx-doc/sphinx/issues/8880
    'sphinx>=2.1.2, <3.5.0',
    'sphinx-airflow-theme',
    'sphinx-argparse>=0.1.13',
    'sphinx-autoapi==1.0.0',
    'sphinx-copybutton',
    'sphinx-jinja~=1.1',
    'sphinx-rtd-theme>=0.1.6',
    'sphinxcontrib-httpdomain>=1.7.0',
    'sphinxcontrib-redoc>=1.6.0',
    'sphinxcontrib-spelling==7.2.1',
]
docker = [
    'docker',
]
drill = ['sqlalchemy-drill>=1.1.0', 'sqlparse>=0.4.1']
druid = [
    'pydruid>=0.4.1',
]
elasticsearch = [
    'elasticsearch>7',
    'elasticsearch-dbapi',
    'elasticsearch-dsl>=5.0.0',
]
exasol = [
    'pyexasol>=0.5.1,<1.0.0',
]
facebook = [
    'facebook-business>=6.0.2',
]
flask_appbuilder_authlib = [
    'authlib',
]
google = [
    'PyOpenSSL',
    'google-ads>=12.0.0',
    'google-api-core>=1.25.1,<2.0.0',
    'google-api-python-client>=1.6.0,<2.0.0',
    'google-auth>=1.0.0,<2.0.0',
    'google-auth-httplib2>=0.0.1',
    'google-cloud-automl>=2.1.0,<3.0.0',
    'google-cloud-bigquery-datatransfer>=3.0.0,<4.0.0',
    'google-cloud-bigtable>=1.0.0,<2.0.0',
    'google-cloud-container>=0.1.1,<2.0.0',
    'google-cloud-datacatalog>=3.0.0,<4.0.0',
    'google-cloud-dataproc>=2.2.0,<3.0.0',
    'google-cloud-dlp>=0.11.0,<2.0.0',
    'google-cloud-kms>=2.0.0,<3.0.0',
    'google-cloud-language>=1.1.1,<2.0.0',
    'google-cloud-logging>=2.1.1,<3.0.0',
    # 1.1.0 removed field_mask and broke import for released providers
    # We can remove the <1.1.0 limitation after we release new Google Provider
    'google-cloud-memcache>=0.2.0,<1.1.0',
    'google-cloud-monitoring>=2.0.0,<3.0.0',
    'google-cloud-os-login>=2.0.0,<3.0.0',
    'google-cloud-pubsub>=2.0.0,<3.0.0',
    'google-cloud-redis>=2.0.0,<3.0.0',
    'google-cloud-secret-manager>=0.2.0,<2.0.0',
    'google-cloud-spanner>=1.10.0,<2.0.0',
    'google-cloud-speech>=0.36.3,<2.0.0',
    'google-cloud-storage>=1.30,<2.0.0',
    'google-cloud-tasks>=2.0.0,<3.0.0',
    'google-cloud-texttospeech>=0.4.0,<2.0.0',
    'google-cloud-translate>=1.5.0,<2.0.0',
    'google-cloud-videointelligence>=1.7.0,<2.0.0',
    'google-cloud-vision>=0.35.2,<2.0.0',
    'google-cloud-workflows>=0.1.0,<2.0.0',
    'grpcio-gcp>=0.2.2',
    'httpx',
    'json-merge-patch~=0.2',
    # pandas-gbq 0.15.0 release broke google provider's bigquery import
    # _check_google_client_version (airflow/providers/google/cloud/hooks/bigquery.py:49)
    'pandas-gbq<0.15.0',
]
grpc = [
    'google-auth>=1.0.0, <2.0.0dev',
    'google-auth-httplib2>=0.0.1',
    'grpcio>=1.15.0',
]
hashicorp = [
    'hvac~=0.10',
]
hdfs = [
    'snakebite-py3',
]
hive = [
    'hmsclient>=0.1.0',
    'pyhive[hive]>=0.6.0;python_version<"3.9"',
    'thrift>=0.9.2',
]
http = [
    # The 2.26.0 release of requests got rid of the chardet LGPL mandatory dependency, allowing us to
    # release it as a requirement for airflow
    'requests>=2.26.0',
]
http_provider = [
    # NOTE ! The HTTP provider is NOT preinstalled by default when Airflow is installed - because it
    #        depends on `requests` library and until `chardet` is mandatory dependency of `requests`
    #        See https://github.com/psf/requests/pull/5797
    #        This means that providers that depend on Http and cannot work without it, have to have
    #        explicit dependency on `apache-airflow-providers-http` which needs to be pulled in for them.
    #        Other cross-provider-dependencies are optional (usually cross-provider dependencies only enable
    #        some features of providers and majority of those providers works). They result with an extra,
    #        not with the `install-requires` dependency.
    'apache-airflow-providers-http',
]
jdbc = [
    'jaydebeapi>=1.1.1',
]
jenkins = [
    'python-jenkins>=1.0.0',
]
jira = [
    'JIRA>1.0.7',
]
kerberos = [
    'pykerberos>=1.1.13',
    'requests_kerberos>=0.10.0',
    'thrift_sasl>=0.2.0',
]
kubernetes = [
    'cryptography>=2.0.0',
    'kubernetes>=3.0.0, <12.0.0',
]
kylin = ['kylinpy>=2.6']
ldap = [
    'ldap3>=2.5.1',
    'python-ldap',
]
leveldb = ['plyvel']
mongo = [
    'dnspython>=1.13.0,<3.0.0',
    'pymongo>=3.6.0',
]
mssql = [
    'pymssql~=2.1,>=2.1.5',
]
mysql = [
    'mysql-connector-python>=8.0.11, <=8.0.22',
    'mysqlclient>=1.3.6,<3',
]
neo4j = ['neo4j>=4.2.1']
odbc = [
    'pyodbc',
]
oracle = [
    'cx_Oracle>=5.1.2',
]
pagerduty = [
    'pdpyras>=4.1.2,<5',
]
papermill = [
    'papermill[all]>=1.2.1',
    'scrapbook[all]',
]
password = [
    'bcrypt>=2.0.0',
    'flask-bcrypt>=0.7.1',
]
pinot = [
    # pinotdb v0.1.1 may still work with older versions of Apache Pinot, but we've confirmed that it
    # causes a problem with newer versions.
    'pinotdb>0.1.2,<1.0.0',
]
plexus = [
    'arrow>=0.16.0',
]
postgres = [
    'psycopg2-binary>=2.7.4',
]
presto = ['presto-python-client>=0.7.0,<0.8']
qubole = [
    'qds-sdk>=1.10.4',
]
rabbitmq = [
    'amqp<5.0.0',
]
redis = [
    'redis~=3.2',
]
salesforce = [
    'simple-salesforce>=1.0.0',
    'tableauserverclient',
]
samba = [
    'smbprotocol>=1.5.0',
]
segment = [
    'analytics-python>=1.2.9',
]
sendgrid = [
    'sendgrid>=6.0.0,<7',
]
sentry = [
    'blinker>=1.1',
    'sentry-sdk>=0.8.0',
]
singularity = ['spython>=0.0.56']
slack = [
    'slack_sdk>=3.0.0,<4.0.0',
]
snowflake = [
    'snowflake-connector-python>=2.4.1',
    'snowflake-sqlalchemy>=1.1.0',
]
spark = [
    'pyspark',
]
ssh = [
    'paramiko>=2.6.0',
    'pysftp>=0.2.9',
    'sshtunnel>=0.1.4,<0.2',
]
statsd = [
    'statsd>=3.3.0, <4.0',
]
tableau = [
    'tableauserverclient',
]
telegram = [
    'python-telegram-bot~=13.0',
]
trino = ['trino']
vertica = [
    'vertica-python>=0.5.1',
]
virtualenv = [
    'virtualenv',
]
webhdfs = [
    'hdfs[avro,dataframe,kerberos]>=2.0.4',
]
winrm = [
    'pywinrm~=0.4',
]
yandex = [
    'yandexcloud>=0.97.0',
]
zendesk = [
    'zdesk',
]
# End dependencies group

devel = [
    'aws_xray_sdk',
    'beautifulsoup4~=4.7.1',
    'black',
    'blinker',
    'bowler',
    'click>=7.1,<9',
    'coverage',
    'filelock',
    'flake8>=3.6.0',
    'flake8-colors',
    'flaky',
    'freezegun',
    'github3.py',
    'gitpython',
    'importlib-resources~=1.4',
    'ipdb',
    'jira',
    'jsondiff',
    'mongomock',
    'moto~=2.0',
    'mypy==0.770',
    'parameterized',
    'paramiko',
    'pipdeptree',
    'pre-commit',
    'pygithub',
    'pysftp',
    'pytest~=6.0',
    'pytest-cov',
    'pytest-instafail',
    'pytest-rerunfailures~=9.1',
    'pytest-timeouts',
    'pytest-xdist',
    'python-jose',
    'pywinrm',
    'qds-sdk>=1.9.6',
    'pytest-httpx',
    'requests_mock',
    'wheel',
    'yamllint',
]

devel_minreq = cgroups + devel + doc + kubernetes + mysql + password
devel_hadoop = devel_minreq + hdfs + hive + kerberos + presto + webhdfs

# Dict of all providers which are part of the Apache Airflow repository together with their requirements
PROVIDERS_REQUIREMENTS: Dict[str, List[str]] = {
    'airbyte': http_provider,
    'amazon': amazon,
    'apache.beam': apache_beam,
    'apache.cassandra': cassandra,
    'apache.drill': drill,
    'apache.druid': druid,
    'apache.hdfs': hdfs,
    'apache.hive': hive,
    'apache.kylin': kylin,
    'apache.livy': http_provider,
    'apache.pig': [],
    'apache.pinot': pinot,
    'apache.spark': spark,
    'apache.sqoop': [],
    'asana': asana,
    'celery': celery,
    'cloudant': cloudant,
    'cncf.kubernetes': kubernetes,
    'databricks': databricks,
    'datadog': datadog,
    'dingding': [],
    'discord': [],
    'docker': docker,
    'elasticsearch': elasticsearch,
    'exasol': exasol,
    'facebook': facebook,
    'ftp': [],
    'google': google,
    'grpc': grpc,
    'hashicorp': hashicorp,
    'http': http,
    'imap': [],
    'jdbc': jdbc,
    'jenkins': jenkins,
    'jira': jira,
    'microsoft.azure': azure,
    'microsoft.mssql': mssql,
    'microsoft.winrm': winrm,
    'mongo': mongo,
    'mysql': mysql,
    'neo4j': neo4j,
    'odbc': odbc,
    'openfaas': [],
    'opsgenie': http_provider,
    'oracle': oracle,
    'pagerduty': pagerduty,
    'papermill': papermill,
    'plexus': plexus,
    'postgres': postgres,
    'presto': presto,
    'qubole': qubole,
    'redis': redis,
    'salesforce': salesforce,
    'samba': samba,
    'segment': segment,
    'sendgrid': sendgrid,
    'sftp': ssh,
    'singularity': singularity,
    'slack': slack,
    'snowflake': snowflake,
    'sqlite': [],
    'ssh': ssh,
    'tableau': tableau,
    'telegram': telegram,
    'trino': trino,
    'vertica': vertica,
    'yandex': yandex,
    'zendesk': zendesk,
}

# Those are all additional extras which do not have their own 'providers'
# The 'apache.atlas' and 'apache.webhdfs' are extras that provide additional libraries
# but they do not have separate providers (yet?), they are merely there to add extra libraries
# That can be used in custom python/bash operators.
ADDITIONAL_EXTRAS_REQUIREMENTS: Dict[str, List[str]] = {
    'apache.atlas': atlas,
    'apache.webhdfs': webhdfs,
}


# Those are extras that are extensions of the 'core' Airflow. They provide additional features
# To airflow core. They do not have separate providers because they do not have any operators/hooks etc.
CORE_EXTRAS_REQUIREMENTS: Dict[str, List[str]] = {
    'async': async_packages,
    'celery': celery,  # also has provider, but it extends the core with the Celery executor
    'cgroups': cgroups,
    'cncf.kubernetes': kubernetes,  # also has provider, but it extends the core with the KubernetesExecutor
    'dask': dask,
    'deprecated_api': deprecated_api,
    'github_enterprise': flask_appbuilder_authlib,
    'google_auth': flask_appbuilder_authlib,
    'kerberos': kerberos,
    'ldap': ldap,
    'leveldb': leveldb,
    'password': password,
    'rabbitmq': rabbitmq,
    'sentry': sentry,
    'statsd': statsd,
    'virtualenv': virtualenv,
}

EXTRAS_REQUIREMENTS: Dict[str, List[str]] = deepcopy(CORE_EXTRAS_REQUIREMENTS)


def add_extras_for_all_providers() -> None:
    """
    Adds extras for all providers.
    By default all providers have the same extra name as provider id, for example
    'apache.hive' extra has 'apache.hive' provider requirement.
    """
    for provider_name, provider_requirement in PROVIDERS_REQUIREMENTS.items():
        EXTRAS_REQUIREMENTS[provider_name] = provider_requirement


def add_additional_extras() -> None:
    """Adds extras for all additional extras."""
    for extra_name, extra_requirement in ADDITIONAL_EXTRAS_REQUIREMENTS.items():
        EXTRAS_REQUIREMENTS[extra_name] = extra_requirement


add_extras_for_all_providers()
add_additional_extras()

#############################################################################################################
#  The whole section can be removed in Airflow 3.0 as those old aliases are deprecated in 2.* series
#############################################################################################################

# Dictionary of aliases from 1.10 - deprecated in Airflow 2.*
EXTRAS_DEPRECATED_ALIASES: Dict[str, str] = {
    'atlas': 'apache.atlas',
    'aws': 'amazon',
    'azure': 'microsoft.azure',
    'cassandra': 'apache.cassandra',
    'crypto': '',  # All crypto requirements are installation requirements of core Airflow
    'druid': 'apache.druid',
    'gcp': 'google',
    'gcp_api': 'google',
    'hdfs': 'apache.hdfs',
    'hive': 'apache.hive',
    'kubernetes': 'cncf.kubernetes',
    'mssql': 'microsoft.mssql',
    'pinot': 'apache.pinot',
    'qds': 'qubole',
    's3': 'amazon',
    'spark': 'apache.spark',
    'webhdfs': 'apache.webhdfs',
    'winrm': 'microsoft.winrm',
}

EXTRAS_DEPRECATED_ALIASES_NOT_PROVIDERS: List[str] = [
    "crypto",
    "webhdfs",
]


def add_extras_for_all_deprecated_aliases() -> None:
    """
    Add extras for all deprecated aliases. Requirements for those deprecated aliases are the same
    as the extras they are replaced with.
    The requirements are not copies - those are the same lists as for the new extras. This is intended.
    Thanks to that if the original extras are later extended with providers, aliases are extended as well.
    """
    for alias, extra in EXTRAS_DEPRECATED_ALIASES.items():
        requirements = EXTRAS_REQUIREMENTS.get(extra) if extra != '' else []
        if requirements is None:
            raise Exception(f"The extra {extra} is missing for deprecated alias {alias}")
        EXTRAS_REQUIREMENTS[alias] = requirements


def add_all_deprecated_provider_packages() -> None:
    """
    For deprecated aliases that are providers, we will swap the providers requirements to instead
    be the provider itself.

    e.g. {"kubernetes": ["kubernetes>=3.0.0, <12.0.0", ...]} becomes
    {"kubernetes": ["apache-airflow-provider-cncf-kubernetes"]}
    """
    for alias, provider in EXTRAS_DEPRECATED_ALIASES.items():
        if alias in EXTRAS_DEPRECATED_ALIASES_NOT_PROVIDERS:
            continue
        replace_extra_requirement_with_provider_packages(alias, [provider])


add_extras_for_all_deprecated_aliases()

#############################################################################################################
#  End of deprecated section
#############################################################################################################

# This is list of all providers. It's a shortcut for anyone who would like to easily get list of
# All providers. It is used by pre-commits.
ALL_PROVIDERS = list(PROVIDERS_REQUIREMENTS.keys())

ALL_DB_PROVIDERS = [
    'apache.cassandra',
    'apache.drill',
    'apache.druid',
    'apache.hdfs',
    'apache.hive',
    'apache.pinot',
    'cloudant',
    'exasol',
    'microsoft.mssql',
    'mongo',
    'mysql',
    'neo4j',
    'postgres',
    'presto',
    'trino',
    'vertica',
]

# Special requirements for all database-related providers. They are de-duplicated.
all_dbs = list({req for db_provider in ALL_DB_PROVIDERS for req in PROVIDERS_REQUIREMENTS[db_provider]})

# Requirements for all "user" extras (no devel). They are de-duplicated. Note that we do not need
# to separately add providers requirements - they have been already added as 'providers' extras above
_all_requirements = list({req for extras_reqs in EXTRAS_REQUIREMENTS.values() for req in extras_reqs})

# All user extras here
EXTRAS_REQUIREMENTS["all"] = _all_requirements

# All db user extras here
EXTRAS_REQUIREMENTS["all_dbs"] = all_dbs

# This can be simplified to devel_hadoop + _all_requirements due to inclusions
# but we keep it for explicit sake. We are de-duplicating it anyway.
devel_all = list(set(_all_requirements + doc + devel_minreq + devel_hadoop))

# Those are packages excluded for "all" dependencies
PACKAGES_EXCLUDED_FOR_ALL = []
PACKAGES_EXCLUDED_FOR_ALL.extend(
    [
        'snakebite',
    ]
)


def is_package_excluded(package: str, exclusion_list: List[str]) -> bool:
    """
    Checks if package should be excluded.

    :param package: package name (beginning of it)
    :param exclusion_list: list of excluded packages
    :return: true if package should be excluded
    """
    return any(package.startswith(excluded_package) for excluded_package in exclusion_list)


devel_all = [
    package
    for package in devel_all
    if not is_package_excluded(package=package, exclusion_list=PACKAGES_EXCLUDED_FOR_ALL)
]

devel_ci = devel_all


# Those are extras that we have to add for development purposes
# They can be use to install some predefined set of dependencies.
EXTRAS_REQUIREMENTS["doc"] = doc
EXTRAS_REQUIREMENTS["devel"] = devel_minreq  # devel_minreq already includes doc
EXTRAS_REQUIREMENTS["devel_hadoop"] = devel_hadoop  # devel_hadoop already includes devel_minreq
EXTRAS_REQUIREMENTS["devel_all"] = devel_all
EXTRAS_REQUIREMENTS["devel_ci"] = devel_ci


def sort_extras_requirements() -> Dict[str, List[str]]:
    """
    For Python 3.6+ the dictionary order remains when keys() are retrieved.
    Sort both: extras and list of dependencies to make it easier to analyse problems
    external packages will be first, then if providers are added they are added at the end of the lists.
    """
    sorted_requirements = dict(sorted(EXTRAS_REQUIREMENTS.items()))
    for extra_list in sorted_requirements.values():
        extra_list.sort()
    return sorted_requirements


EXTRAS_REQUIREMENTS = sort_extras_requirements()

# Those providers are pre-installed always when airflow is installed.
# Those providers do not have dependency on airflow2.0 because that would lead to circular dependencies.
# This is not a problem for PIP but some tools (pipdeptree) show those as a warning.
PREINSTALLED_PROVIDERS = [
    'ftp',
    'http',
    'imap',
    'sqlite',
]


def get_provider_package_from_package_id(package_id: str) -> str:
    """
    Builds the name of provider package out of the package id provided/

    :param package_id: id of the package (like amazon or microsoft.azure)
    :return: full name of package in PyPI
    """
    package_suffix = package_id.replace(".", "-")
    return f"apache-airflow-providers-{package_suffix}"


def get_excluded_providers() -> List[str]:
    """
    Returns packages excluded for the current python version.
    Currently the only excluded provider is apache hive for Python 3.9.
    Until https://github.com/dropbox/PyHive/issues/380 is fixed.
    """
    return ['apache.hive'] if PY39 else []


def get_all_provider_packages() -> str:
    """Returns all provider packages configured in setup.py"""
    excluded_providers = get_excluded_providers()
    return " ".join(
        get_provider_package_from_package_id(package)
        for package in PROVIDERS_REQUIREMENTS
        if package not in excluded_providers
    )


class AirflowDistribution(Distribution):
    """The setuptools.Distribution subclass with Airflow specific behaviour"""

    def parse_config_files(self, *args, **kwargs) -> None:
        """
        Ensure that when we have been asked to install providers from sources
        that we don't *also* try to install those providers from PyPI.
        Also we should make sure that in this case we copy provider.yaml files so that
        Providers manager can find package information.
        """
        super().parse_config_files(*args, **kwargs)
        if os.getenv(INSTALL_PROVIDERS_FROM_SOURCES) == 'true':
            self.install_requires = [
                req for req in self.install_requires if not req.startswith('apache-airflow-providers-')
            ]
            provider_yaml_files = glob.glob("airflow/providers/**/provider.yaml", recursive=True)
            for provider_yaml_file in provider_yaml_files:
                provider_relative_path = relpath(provider_yaml_file, os.path.join(my_dir, "airflow"))
                self.package_data['airflow'].append(provider_relative_path)
        else:
            self.install_requires.extend(
                [get_provider_package_from_package_id(package_id) for package_id in PREINSTALLED_PROVIDERS]
            )


def replace_extra_requirement_with_provider_packages(extra: str, providers: List[str]) -> None:
    """
    Replaces extra requirement with provider package. The intention here is that when
    the provider is added as dependency of extra, there is no need to add the dependencies
    separately. This is not needed and even harmful, because in case of future versions of
    the provider, the requirements might change, so hard-coding requirements from the version
    that was available at the release time might cause dependency conflicts in the future.

    Say for example that you have salesforce provider with those deps:

    { 'salesforce': ['simple-salesforce>=1.0.0', 'tableauserverclient'] }

    Initially ['salesforce'] extra has those requirements and it works like that when you install
    it when INSTALL_PROVIDERS_FROM_SOURCES is set to `true` (during the development). However, when
    the production installation is used, The dependencies are changed:

    { 'salesforce': ['apache-airflow-providers-salesforce'] }

    And then, 'apache-airflow-providers-salesforce' package has those 'install_requires' dependencies:
            ['simple-salesforce>=1.0.0', 'tableauserverclient']

    So transitively 'salesforce' extra has all the requirements it needs and in case the provider
    changes it's dependencies, they will transitively change as well.

    In the constraint mechanism we save both - provider versions and it's dependencies
    version, which means that installation using constraints is repeatable.

    :param extra: Name of the extra to add providers to
    :param providers: list of provider ids
    """
    EXTRAS_REQUIREMENTS[extra] = [
        get_provider_package_from_package_id(package_name) for package_name in providers
    ]


def add_provider_packages_to_extra_requirements(extra: str, providers: List[str]) -> None:
    """
    Adds provider packages as requirements to extra. This is used to add provider packages as requirements
    to the "bulk" kind of extras. Those bulk extras do not have the detailed 'extra' requirements as
    initial values, so instead of replacing them (see previous function) we can extend them.

    :param extra: Name of the extra to add providers to
    :param providers: list of provider ids
    """
    EXTRAS_REQUIREMENTS[extra].extend(
        [get_provider_package_from_package_id(package_name) for package_name in providers]
    )


def add_all_provider_packages() -> None:
    """
    In case of regular installation (providers installed from packages), we should add extra dependencies to
    Airflow - to get the providers automatically installed when those extras are installed.

    For providers installed from sources we skip that step. That helps to test and install airflow with
    all packages in CI - for example when new providers are added, otherwise the installation would fail
    as the new provider is not yet in PyPI.

    """
    for provider in ALL_PROVIDERS:
        replace_extra_requirement_with_provider_packages(provider, [provider])
    add_provider_packages_to_extra_requirements("all", ALL_PROVIDERS)
    add_provider_packages_to_extra_requirements("devel_ci", ALL_PROVIDERS)
    add_provider_packages_to_extra_requirements("devel_all", ALL_PROVIDERS)
    add_provider_packages_to_extra_requirements("all_dbs", ALL_DB_PROVIDERS)
    add_provider_packages_to_extra_requirements(
        "devel_hadoop", ["apache.hdfs", "apache.hive", "presto", "trino"]
    )
    add_all_deprecated_provider_packages()


class Develop(develop_orig):
    """Forces removal of providers in editable mode."""

    def run(self) -> None:
        self.announce('Installing in editable mode. Uninstalling provider packages!', level=log.INFO)
        # We need to run "python3 -m pip" because it might be that older PIP binary is in the path
        # And it results with an error when running pip directly (cannot import pip module)
        # also PIP does not have a stable API so we have to run subprocesses ¯\_(ツ)_/¯
        try:
            installed_packages = (
                subprocess.check_output(["python3", "-m", "pip", "freeze"]).decode().splitlines()
            )
            airflow_provider_packages = [
                package_line.split("=")[0]
                for package_line in installed_packages
                if package_line.startswith("apache-airflow-providers")
            ]
            self.announce(f'Uninstalling ${airflow_provider_packages}!', level=log.INFO)
            subprocess.check_call(["python3", "-m", "pip", "uninstall", "--yes", *airflow_provider_packages])
        except subprocess.CalledProcessError as e:
            self.announce(f'Error when uninstalling airflow provider packages: {e}!', level=log.WARN)
        super().run()


class Install(install_orig):
    """Forces installation of providers from sources in editable mode."""

    def run(self) -> None:
        self.announce('Standard installation. Providers are installed from packages', level=log.INFO)
        super().run()


def do_setup() -> None:
    """
    Perform the Airflow package setup.

    Most values come from setup.cfg, only the dynamically calculated ones are passed to setup
    function call. See https://setuptools.readthedocs.io/en/latest/userguide/declarative_config.html
    """
    setup_kwargs = {}

    def include_provider_namespace_packages_when_installing_from_sources() -> None:
        """
        When installing providers from sources we install all namespace packages found below airflow,
        including airflow and provider packages, otherwise defaults from setup.cfg control this.
        The kwargs in setup() call override those that are specified in setup.cfg.
        """
        if os.getenv(INSTALL_PROVIDERS_FROM_SOURCES) == 'true':
            setup_kwargs['packages'] = find_namespace_packages(include=['airflow*'])

    include_provider_namespace_packages_when_installing_from_sources()
    if os.getenv(INSTALL_PROVIDERS_FROM_SOURCES) == 'true':
        print("Installing providers from sources. Skip adding providers as dependencies")
    else:
        add_all_provider_packages()

    write_version()
    setup(
        distclass=AirflowDistribution,
        version=version,
        # extras_require=EXTRAS_REQUIREMENTS,
        # download_url=('https://archive.apache.org/dist/airflow/' + version),
        cmdclass={
            'extra_clean': CleanCommand,
            'compile_assets': CompileAssets,
            'list_extras': ListExtras,
            'install': Install,
            'develop': Develop,
        },
        test_suite='setup.airflow_test_suite',
        **setup_kwargs,
    )


if __name__ == "__main__":
    do_setup()
